"""Meteocat client."""

from collections.abc import Mapping

import pandas as pd
import pyproj

from meteora import settings
from meteora.clients.base import BaseJSONClient
from meteora.mixins import (
    APIKeyHeaderMixin,
    StationsEndpointMixin,
    VariablesEndpointMixin,
)
from meteora.utils import DateTimeType, KwargsType, RegionType, VariablesType

# API endpoints
BASE_URL = "https://api.meteo.cat/xema/v1"
STATIONS_ENDPOINT = f"{BASE_URL}/estacions/metadades"
VARIABLES_ENDPOINT = f"{BASE_URL}/variables/mesurades/metadades"
TS_ENDPOINT = f"{BASE_URL}/variables/mesurades"

# useful constants
STATIONS_GDF_ID_COL = "codi"
TS_DF_STATIONS_ID_COL = "codi"
TS_DF_TIME_COL = "data"
# VARIABLES_NAME_COL = "nom"
VARIABLES_ID_COL = "codi"
ECV_DICT = {
    # precipitation
    "precipitation": 35,  # "Precipitació",
    # pressure
    "pressure": 34,  # "Pressió atmosfèrica",
    # radiation budget
    "radiation_shortwave": 39,  # "Radiació UV",
    # temperature
    "temperature": 32,  # "Temperatura",
    # water vapour
    "relative_humidity": 33,  # "Humitat relativa",
    # wind
    "wind_speed": 46,  # "Velocitat del vent a 2 m (esc.)",
    "wind_direction": 47,  # "Direcció de vent 2 m (m. 1)",
}


class MeteocatClient(
    APIKeyHeaderMixin,
    StationsEndpointMixin,
    VariablesEndpointMixin,
    BaseJSONClient,
):
    """Meteocat client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    api_key : str
        Meteocat API key.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # geom constants
    X_COL = "coordenades.longitud"
    Y_COL = "coordenades.latitud"
    CRS = pyproj.CRS("epsg:4326")

    # API endpoints
    _stations_endpoint = STATIONS_ENDPOINT
    _variables_endpoint = VARIABLES_ENDPOINT
    _ts_endpoint = TS_ENDPOINT

    # data frame labels constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _ts_df_time_col = TS_DF_TIME_COL
    # _variables_name_col = VARIABLES_NAME_COL
    _variables_id_col = VARIABLES_ID_COL
    _ecv_dict = ECV_DICT

    def __init__(
        self, region: RegionType, api_key: str, **sjoin_kwargs: KwargsType
    ) -> None:
        """Initialize Meteocat client."""
        self.region = region
        self._api_key = api_key
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs

        # need to call super().__init__() to set the cache
        super().__init__()

    def _stations_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        return pd.json_normalize(response_content)

    def _variables_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        return pd.json_normalize(response_content)

    def _ts_df_from_content(self, response_content: Mapping):
        # process response
        response_df = pd.json_normalize(response_content)
        # filter stations
        response_df = response_df[response_df["codi"].isin(self.stations_gdf.index)]
        # extract json observed data, i.e.,  the "variables" column into a list of data
        # frames and concatenate them into a single data frame
        ts_df = pd.concat(
            response_df.apply(
                lambda row: pd.DataFrame(row["variables"][0]["lectures"]), axis=1
            ).tolist()
        )
        # add the station id column matching the observations
        ts_df[self._ts_df_stations_id_col] = (
            response_df[self._ts_df_stations_id_col]
            .repeat(
                response_df.apply(
                    lambda row: len(row["variables"][0]["lectures"]), axis=1
                )
            )
            .values
        )
        # TODO: values_col as class-level constant?
        values_col = "valor"
        # # convert to a wide data frame
        # ts_df = long_df.pivot_table(
        #     index=self._time_col, columns=self._stations_id_col, values=values_col
        # )
        # # set the index name
        # ts_df.index.name = settings.TIME_NAME
        # # convert the index from string to datetime
        # ts_df.index = pd.to_datetime(ts_df.index)
        # ACHTUNG: do not sort the index here
        # note that we are renaming a series
        return ts_df.assign(
            **{self._ts_df_time_col: pd.to_datetime(ts_df[self._ts_df_time_col])}
        ).set_index([self._ts_df_stations_id_col, self._ts_df_time_col])[values_col]

    # def _get_date_ts_df(
    #     self,
    #     variable_id: int,
    #     date: datetime.date,
    # ) -> pd.DataFrame:
    #     """Get time series data frame for a given day.

    #     Parameters
    #     ----------
    #     variable_id : int
    #         Meteocat variable code.
    #     date : datetime.date
    #         datetime.date instance for the requested data period.

    #     Returns
    #     -------
    #     ts_df : pd.DataFrame
    #         Data frame with a time series of measurements (rows) at each station
    #         (columns).

    #     """
    #     # # process date arg
    #     # if isinstance(date, str):
    #     #     date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
    #     # request url
    #     request_url = (
    #         f"{self._ts_endpoint}"
    #         f"/{variable_id}/{date.year}/{date.month:02}/{date.day:02}"
    #     )
    #     response_content = self._get_content_from_url(request_url)
    #     return self._ts_df_from_content(response_content).rename(variable_id)

    def _ts_df_from_endpoint(self, ts_params: Mapping) -> pd.DataFrame:
        # the API only allows returning data for a given day and variable so we have to
        # iterate over the date range and variables to obtain data for all days
        date_range = pd.date_range(
            start=ts_params["start"], end=ts_params["end"], freq="D"
        )
        return pd.concat(
            [
                pd.concat(
                    [
                        # self._get_date_ts_df(variable_id, date)
                        self._ts_df_from_content(
                            self._get_content_from_url(
                                f"{self._ts_endpoint}/{variable_id}/"
                                f"{date.year}/{date.month:02}/{date.day:02}"
                            )
                        ).rename(variable_id)
                        for variable_id in ts_params["variable_ids"]
                    ],
                    axis="columns",
                    ignore_index=False,
                )
                for date in date_range
            ],
            axis="index",
            ignore_index=False,
        )

    def get_ts_df(
        self,
        variables: VariablesType,
        start: DateTimeType,
        end: DateTimeType,
    ) -> pd.DataFrame:
        """Get time series data frame.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either a Meteocat variable code (integer or
            string) or an essential climate variable (ECV) following the Meteora
            nomenclature (string).
        start, end : datetime-like, str, int, float
            Values representing the start and end of the requested data period
            respectively. Accepts any datetime-like object that can be passed to
            pandas.Timestamp.

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        ts_df = self._get_ts_df(
            variables,
            start=start,
            end=end,
        )
        # filter time range to avoid including a full day after
        # TODO: dry with Agrometeo, perhaps a global approach in the base client
        time_ser = ts_df.index.get_level_values(settings.TIME_COL).to_series()
        tz = time_ser.dt.tz
        return ts_df.loc[
            (
                slice(None),
                time_ser.between(
                    pd.Timestamp(start, tz=tz),
                    pd.Timestamp(end, tz=tz),
                    inclusive="both",
                ),
            ),
            :,
        ]



"""Stations mixins."""

from abc import ABC

import pandas as pd

from meteora import utils


class StationsEndpointMixin(ABC):
    """Stations endpoint mixin."""

    @utils.abstract_attribute
    def _stations_endpoint(self) -> str:
        pass

    def _get_stations_df(self) -> pd.DataFrame:
        """Get the stations dataframe for the instance.

        Returns
        -------
        stations_df : pandas.DataFrame
            The stations data for the given region.

        """
        response_content = self._get_content_from_url(self._stations_endpoint)
        return self._stations_df_from_content(response_content)



"""National Oceanic And Atmospheric Administration (NOAA) client."""

import logging as lg
from collections.abc import Sequence

import dask
import pandas as pd
import pooch
import pyproj
import requests
from dask import diagnostics

from meteora import settings, utils
from meteora.clients.base import BaseTextClient
from meteora.mixins import StationsEndpointMixin, VariablesHardcodedMixin
from meteora.utils import DateTimeType, KwargsType, RegionType, VariablesType

# disable pooch warnings when providing `None` as "known_hash"
logger = pooch.get_logger()
logger.setLevel("WARNING")


# API endpoints
BASE_URL = "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly"
GHCNH_STATIONS_ENDPOINT = f"{BASE_URL}/doc/ghcnh-station-list.csv"
TS_ENDPOINT = f"{BASE_URL}/access/by-year/" + "{year}/psv/GHCNh_{station_id}_{year}.psv"

# for pooch
# TODO: how often can it change? how to properly update it if it does?
STATIONS_LIST_KNOWN_HASH = (
    "6b837d8d953aa6ebc172755864e549518944e7bec15b2196ee191219587a96f5"
)

# useful constants
STATIONS_GDF_ID_COL = "GHCN_ID"
# ACHTUNG: note that in the time series data frame the station column label is "Station
# ID" whereas in the stations data frame it is "id".
TS_DF_STATIONS_ID_COL = "STATION"
TS_DF_TIME_COL = "DATE"
VARIABLES_ID_COL = "code"
VARIABLES_LABEL_COL = "description"

# see section "IV. List of elements/variable" and appendix A of the GHCNh documentation
# www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/doc/
# ghcnh_DOCUMENTATION.pdf
VARIABLES_DICT = {
    "temperature": "2 meter (circa) Above Ground Level Air (dry bulb) Temperature (⁰C n"
    "to tenths)",
    "relative_humidity": "Depending on the source, relative humidity is either measured"
    " directly or calculated from air (dry bulb) temperature and dew point temperature "
    "(whole percent)",
    "station_level_pressure": "The pressure that is observed at a specific elevation "
    "and is the true barometric pressure of a location. It is the pressure exerted by "
    "the atmosphere at a point as a result of gravity acting upon the 'column' of air "
    "that lies directly above the point. (hPa)",
    "precipitation": "Total liquid precipitation (rain or melted snow). Totals are "
    "nominally for the hour, but may include intermediate reports within the hour. "
    "Please refer to Appendix B for important details on precipitation totals; a `T` in"
    " the measurement code column indicates a trace amount of precipitation "
    "(millimeters).",
    "wind_speed": "Wind speed (meters per second)",
    "wind_direction": "Wind Direction from true north using compass directions (e.g. "
    "360 = true north, 180 = south, 270 = west, etc.). Note: A direction of `000` is "
    "given for calm winds. (whole degrees)",
}
ECV_DICT = {
    # precipitation
    "preciptation": "precipitation",
    # pressure
    "pressure": "station_level_pressure",
    # temperature
    "temperature": "temperature",
    # water vapour
    "relative_humidity": "relative_humidity",
    # wind
    "wind_speed": "wind_speed",
    "wind_direction": "wind_direction",
}


class GHCNHourlyClient(StationsEndpointMixin, VariablesHardcodedMixin, BaseTextClient):
    """NOAA GHCN hourly client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    pooch_kwargs : dict, optional
        Keyword arguments to pass to the `pooch.retrieve` function when downloading the
        stations time series data.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # ACHTUNG: many constants are set in `GHCNH_STATIONS_COLUMNS` above
    # geom constants
    X_COL = "LONGITUDE"
    Y_COL = "LATITUDE"
    CRS = pyproj.CRS("epsg:4326")

    # API endpoints
    _stations_endpoint = GHCNH_STATIONS_ENDPOINT
    _ts_endpoint = TS_ENDPOINT

    # data frame labels constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _ts_df_time_col = TS_DF_TIME_COL
    _variables_id_col = VARIABLES_ID_COL
    _variables_label_col = VARIABLES_LABEL_COL
    _variables_dict = VARIABLES_DICT
    _ecv_dict = ECV_DICT

    def __init__(
        self,
        region: RegionType,
        *,
        pooch_kwargs: KwargsType | None = None,
        **sjoin_kwargs: KwargsType,
    ) -> None:
        """Initialize GHCN hourly client."""
        self.region = region
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs

        if pooch_kwargs is None:
            pooch_kwargs = {}
        self.pooch_kwargs = pooch_kwargs

        # need to call super().__init__() to set the cache
        super().__init__()

    def _get_stations_df(self) -> pd.DataFrame:
        """Get all stations."""
        try:
            # ACHTUNG: we are not using `self.pooch_kwargs` to set the path to download
            # files, because the stations file path can be reused independently of the
            # region, so it is useful to cache it globally
            stations_filepath = pooch.retrieve(
                self._stations_endpoint, STATIONS_LIST_KNOWN_HASH
            )
        except ValueError:
            # sha hash mismatch, probably because of updates in the station list
            stations_filepath = pooch.retrieve(self._stations_endpoint, None)
        return pd.read_csv(stations_filepath)

    def _ts_params(
        self, variable_ids: Sequence, start: DateTimeType, end: DateTimeType
    ) -> dict:
        # process date args
        start = pd.Timestamp(start)
        end = pd.Timestamp(end)

        return dict(variable_ids=variable_ids, start=start, end=end)

    def _ts_df_from_endpoint(self, ts_params) -> pd.DataFrame:
        """Get time series data frame from endpoint."""
        # we override this method because we need a separate request for each station
        variable_cols = list(ts_params["variable_ids"])
        cols_to_keep = (
            [self._ts_df_stations_id_col] + [self._ts_df_time_col] + variable_cols
        )
        start = ts_params["start"]
        end = ts_params["end"]
        requested_years = set(range(start.year, end.year + 1))

        # def _process_station_ts_df(year, station_id):
        #     try:
        #         station_ts_filepath = pooch.retrieve(
        #             self._ts_endpoint.format(year=year, station_id=station_id),
        #             None,
        #             **self.pooch_kwargs,
        #         )
        #     except requests.HTTPError:
        #         return pd.DataFrame()
        #     ts_df = pd.read_csv(station_ts_filepath, sep="|")[cols_to_keep]
        #     ts_df[self._ts_df_time_col] = pd.to_datetime(ts_df[self._ts_df_time_col])
        #     return ts_df[ts_df[self._ts_df_time_col].between(start, end)]

        # ts_df = pd.concat(
        #     [
        #         _process_station_ts_df(year, station_id)
        #         for station_id in self.stations_gdf.index
        #         for year in requested_years
        #     ]
        # )

        # use dask to parallelize requests
        def _process_station_ts_df(year, station_id):
            try:
                station_ts_filepath = pooch.retrieve(
                    self._ts_endpoint.format(year=year, station_id=station_id),
                    None,
                    **self.pooch_kwargs,
                )
            except requests.HTTPError:
                # TODO: log?
                return pd.DataFrame()
            ts_df = pd.read_csv(station_ts_filepath, sep="|", usecols=cols_to_keep)
            ts_df[self._ts_df_time_col] = pd.to_datetime(ts_df[self._ts_df_time_col])
            return ts_df[ts_df[self._ts_df_time_col].between(start, end)]

        tasks = [
            dask.delayed(_process_station_ts_df)(year, station_id)
            for station_id in self.stations_gdf.index
            for year in requested_years
        ]
        with diagnostics.ProgressBar():
            ts_dfs = dask.compute(*tasks)

        try:
            return pd.concat([ts_df for ts_df in ts_dfs if not ts_df.empty]).set_index(
                [self._ts_df_stations_id_col, self._ts_df_time_col]
            )
        except ValueError:  # no objects to concatenate
            utils.log(
                "No data found for the requested period and stations, returning an ",
                level=lg.WARNING,
            )
            return pd.DataFrame(columns=variable_cols)

    def _post_process_ts_df(self, ts_df: pd.DataFrame) -> pd.DataFrame:
        # no need to sort the index given the way the data has been requested
        return ts_df.apply(pd.to_numeric, axis="columns")

    def get_ts_df(
        self,
        variables: VariablesType,
        start: DateTimeType,
        end: DateTimeType,
    ) -> pd.DataFrame:
        """Get time series data frame.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either an GHCNh variable code (string) or an
            essential climate variable (ECV) following the Meteora nomenclature
            (string).
        start, end : datetime-like, str, int, float
            Values representing the start and end of the requested data period
            respectively. Accepts any datetime-like object that can be passed to
            pandas.Timestamp.

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        return self._get_ts_df(variables, start, end)



"""MeteoSwiss client."""

import datetime as dt
from collections.abc import Mapping

import pandas as pd
import pyproj

from meteora import settings
from meteora.clients.base import BaseTextClient
from meteora.mixins import StationsEndpointMixin, VariablesEndpointMixin
from meteora.utils import CRSType, DateTimeType, KwargsType, RegionType, VariablesType

BASE_URL = "https://data.geo.admin.ch/ch.meteoschweiz.ogd-smn"
STATIONS_ENDPOINT = f"{BASE_URL}/ogd-smn_meta_stations.csv"
VARIABLES_ENDPOINT = f"{BASE_URL}/ogd-smn_meta_parameters.csv"
# there are several ways to structure a time series request, for more details see
# https://opendatadocs.meteoswiss.ch/general/download#how-csv-files-are-structured
# - the `t` corresponds to the "original data" (Originalwert), i.e., the 10min value,
#   all the other granularities are aggregations (e.g., `h` for hourly...)
# - the `{update_freq}` part can be either `historical` (until Dec 31st of last year) or
#   `recent` (from Jan 1st of this year until yesterday).
# Additionally, for `historical` data, an additional part of the form `_{decade}` must
# be added to the URL, e.g., `_1980-1989` for the 1980s decade.
TS_ENDPOINT = f"{BASE_URL}/" + "{station_id}/ogd-smn_{station_id}_t_{update_freq}.csv"

# useful constants
# TODO: DRY with Agrometeo?
LONLAT_CRS = pyproj.CRS("epsg:4326")
LV95_CRS = pyproj.CRS("epsg:2056")
# ACHTUNG: for some reason, the API mixes up the longitude and latitude columns ONLY in
# the CH1903/LV03 projection. This is why we need to swap the columns in the dict below.
GEOM_COL_DICT = {
    LONLAT_CRS: ["station_coordinates_wgs84_lat", "station_coordinates_wgs84_lon"],
    LV95_CRS: ["station_coordinates_lv95_east", "station_coordinates_lv95_north"],
}
DEFAULT_CRS = LV95_CRS
READ_CSV_KWARGS = dict(sep=";", encoding="ISO-8859-1")
# stations column used by the MeteoSwiss API (do not change)
STATIONS_GDF_ID_COL = "station_abbr"
TS_DF_STATIONS_ID_COL = "station_abbr"
TS_DF_TIME_COL = "reference_timestamp"
VARIABLES_ID_COL = "parameter_shortname"
ECV_DICT = {
    # precipitation
    # "Precipitation (ten minutes total) [mm]"
    "precipitation": "rre150z0",
    # pressure
    # "Atmospheric pressure at barometric altitude (current value) [hPa]"
    "pressure": "prestas0",
    # radiation budget
    # "Global radiation (ten minutes mean) [W/m2]"
    "radiation_shortwave": "gre000z0",
    # "Longwave incoming radiation; ten minute mean [W/m2]"
    "radiation_longwave_incoming": "oli000z0",
    # "Longwave outgoing radiation; ten minute mean [W/m2]"
    "radiation_longwave_outgoing": "olo000z0",
    # temperature
    # "Air temperature 2 m above ground (current value) [°C]"
    "temperature": "tre200s0",
    # water vapour
    # "Dew point 2 m above ground [°C]"
    "dew_point": "tde200h0",
    # "Relative air humidity 2 m above ground (current value) [%]"
    "relative_humidity": "ure200s0",
    # wind
    # "Wind speed scalar (ten minutes mean) [m/s]",
    "wind_speed": "fkl010z0",
    # "Wind direction (ten minutes mean) [degrees]",
    "wind_direction": "dkl010z0",
}


class MeteoSwissClient(StationsEndpointMixin, VariablesEndpointMixin, BaseTextClient):
    """MeteoSwiss client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    crs : str, dict or pyproj.CRS, optional
        The coordinate reference system (CRS) to be used. For Agrometeo, the
        provided value must be equivalent to either the EPSG:21781 (default) or
        EPSG:4326.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # API endpoints
    _stations_endpoint = STATIONS_ENDPOINT
    _variables_endpoint = VARIABLES_ENDPOINT
    _ts_endpoint = TS_ENDPOINT

    # data frame labels constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _ts_df_time_col = TS_DF_TIME_COL
    _variables_id_col = VARIABLES_ID_COL
    _ecv_dict = ECV_DICT

    def __init__(
        self,
        region: RegionType,
        *,
        crs: CRSType | None = None,
        **sjoin_kwargs: KwargsType,
    ) -> None:
        """Initialize MeteoSwiss client."""
        # TODO: DRY with Agrometeo?
        if crs is not None:
            crs = pyproj.CRS(crs)
        else:
            crs = DEFAULT_CRS
        self.CRS = crs
        # self._variables_name_col = variables_name_col or VARIABLES_NAME_COL
        try:
            self.X_COL, self.Y_COL = GEOM_COL_DICT[self.CRS]
        except KeyError:
            raise ValueError(
                f"CRS must be among {list(GEOM_COL_DICT.keys())}, got {self.CRS}"
            )

        self.region = region
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs

        # need to call super().__init__() to set the cache
        super().__init__()

    def _stations_df_from_content(self, response_content) -> pd.DataFrame:
        # return pd.read_csv(self._stations_endpoint, encoding=FILE_ENCODING)
        return pd.read_csv(response_content, **READ_CSV_KWARGS)

    def _variables_df_from_content(self, response_content) -> pd.DataFrame:
        return pd.read_csv(response_content, **READ_CSV_KWARGS)

    def _ts_df_from_content(self, response_content) -> pd.DataFrame:
        return pd.read_csv(response_content, **READ_CSV_KWARGS)

    def _ts_df_from_endpoint(self, ts_params: Mapping) -> pd.DataFrame:
        # the API only allows returning data for a given station so we have to iterate
        # over the list of stations
        # determine whether we need to access the "historical" or "recent" URL, see
        # https://opendatadocs.meteoswiss.ch/general/download#update-frequency
        this_year_start = dt.date(dt.datetime.now().year, 1, 1)

        # TODO: better approach to ensure datetime types in `ts_params`
        start = pd.Timestamp(ts_params["start"])
        if start.date() > this_year_start:
            # the first requested data is after the start of the current year, so we
            # only need to query the data from the "recent" file
            def _get_station_urls(station_id):
                return [
                    self._ts_endpoint.format(
                        station_id=station_id,
                        update_freq="recent",
                    )
                ]
        else:
            # TODO: better approach to ensure datetime types in `ts_params`
            end = pd.Timestamp(ts_params["end"])
            # we need to query the data from at least one "historical" file (there is a
            # "historical" file for each decade, e.g., 1980-1989, 1990-1999, etc.)
            decades = []
            for year in range((start.year // 10) * 10, (end.year // 10) * 10 + 1, 10):
                decades.append(f"{year}-{year + 9}")

            def _get_station_decades_urls(station_id):
                return [
                    self._ts_endpoint.format(
                        station_id=station_id,
                        update_freq=f"historical_{decade}",
                    )
                    for decade in decades
                ]

            if end.date() < this_year_start:
                # the last requested data is before the start of the current year, so we
                # only need to query the data from the "historical" file of each decade
                def _get_station_urls(station_id):
                    return _get_station_decades_urls(station_id)
            else:
                # assume that we need to query both the "historical" and "recent" files
                def _get_station_urls(station_id):
                    return _get_station_decades_urls(station_id) + [
                        self._ts_endpoint.format(
                            station_id=station_id,
                            update_freq="recent",
                        )
                    ]

        def _ts_df_from_url(station_url):
            # read file into a pandas data frame
            ts_df = self._ts_df_from_content(self._get_content_from_url(station_url))
            # set time and station id columns as index
            ts_df = ts_df.assign(
                **{
                    self._ts_df_time_col: pd.to_datetime(
                        ts_df[self._ts_df_time_col], format="%d.%m.%Y %H:%M"
                    )
                }
            ).set_index([self._ts_df_stations_id_col, self._ts_df_time_col])
            # filter time range
            # TODO: dry with Agrometeo and Meteocat
            time_ser = ts_df.index.get_level_values(self._ts_df_time_col).to_series()
            tz = time_ser.dt.tz
            return ts_df.loc[
                (
                    slice(None),
                    time_ser.between(
                        pd.Timestamp(start, tz=tz),
                        pd.Timestamp(end, tz=tz),
                        inclusive="both",
                    ),
                ),
                :,
            ]

        return pd.concat(
            [
                _ts_df_from_url(station_url)
                for station_id in self.stations_gdf.index.str.lower()
                for station_url in _get_station_urls(station_id)
            ],
            axis="index",
        )

    def get_ts_df(
        self,
        variables: VariablesType,
        start: DateTimeType,
        end: DateTimeType,
    ) -> pd.DataFrame:
        """Get time series data frame.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either an Agrometeo variable code (integer or
            string) or an essential climate variable (ECV) following the Meteora
            nomenclature (string).
        start, end : datetime-like, str, int, float
            Values representing the start and end of the requested data period
            respectively. Accepts any datetime-like object that can be passed to
            pandas.Timestamp.

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        return self._get_ts_df(
            variables=variables,
            start=start,
            end=end,
        )



"""Utils.

Based on osmnx utils and downloader modules.
"""

import datetime as dt
import logging as lg
import os
import sys
import unicodedata
from collections.abc import Callable, Mapping, Sequence
from contextlib import redirect_stdout
from pathlib import Path
from typing import IO, TypeVar

import geopandas as gpd
import numpy as np
import pandas as pd
from pyproj.crs import CRS

from meteora import settings

try:
    import xarray as xr
    import xvec  # noqa: F401
except ImportError:
    xr = None

RegionType = str | Sequence | gpd.GeoSeries | gpd.GeoDataFrame | os.PathLike | IO
VariablesType = str | int | list[str] | list[int]
DateTimeType = dt.date | dt.datetime | np.datetime64 | pd.Timestamp | str | int | float
CRSType = str | dict | CRS
KwargsType = Mapping | None
PathType = str | os.PathLike
if xr is not None:
    CubeType = xr.Dataset
else:
    CubeType = TypeVar("CubeType")
AggFuncType = str | Callable | None


########################################################################################
# geo utils
def dms_to_decimal(ser: pd.Series) -> pd.Series:
    """Convert a series from degrees, minutes, seconds (DMS) to decimal degrees."""
    degrees = ser.str[0:2].astype(int)
    minutes = ser.str[2:4].astype(int)
    seconds = ser.str[4:6].astype(int)
    direction = ser.str[-1]

    decimal = degrees + minutes / 60 + seconds / 3600
    decimal = decimal.where(direction.isin(["N", "E"]), -decimal)

    return decimal


########################################################################################
# data structure utils
def long_to_wide(
    ts_df: pd.DataFrame, *, variables: VariablesType | None = None
) -> pd.DataFrame:
    """Convert a time series data frame from long (default) to wide format.

    Parameters
    ----------
    ts_df : pd.DataFrame
        Long form data frame with a time series of measurements (second-level index) at
        each station (first-level index) for each variable (column).
    variables : str, int or list-like of str or int, optional
        Target variables, which must be columns in `ts_df`.

    Returns
    -------
    wide_ts_df : pd.DataFrame
        Wide form data frame with a time series of measurements (index) for each
        variable (first-level column index) at each station (second-level column index).
        If there is only one variable, the column index is a single level featuring the
        stations.
    """
    # despite ruff rule PD010, use unstack which is both simpler and faster
    # https://docs.astral.sh/ruff/rules/pandas-use-of-dot-pivot-or-unstack
    rename_col_level = True
    if variables is None:
        variables = ts_df.columns
    if pd.api.types.is_list_like(variables) and len(variables) == 1:
        variables = variables[0]
        rename_col_level = False
    wide_ts_df = ts_df[variables].unstack(level=ts_df.index.names[0])
    # rename the variables column level (if we have it - i.e., multivariate case)
    if rename_col_level:
        wide_ts_df = wide_ts_df.rename_axis(columns={None: "variable"})
    return wide_ts_df


def long_to_cube(
    ts_df: pd.DataFrame,
    stations_gdf: gpd.GeoDataFrame,
) -> CubeType:
    """Convert a time series data frame and station locations to a vector data cube.

    A vector data cube is an n-D array with at least one dimension indexed by vector
    geometries. In Python, this is represented as an xarray Dataset (or DataArray)
    object with an indexed dimension with vector geometries set using xvec.

    Parameters
    ----------
    ts_df : pd.DataFrame
        Long form data frame with a time series of measurements (second-level index) at
        each station (first-level index) for each variable (column).
    stations_gdf : gpd.GeoDataFrame
        The stations data as a GeoDataFrame.

    Returns
    -------
    ts_cube : xr.Dataset
        The vector data cube with the time series of measurements for each station. The
        stations are indexed by their geometry.
    """
    # get the stations id column in the time series data frame
    stations_ts_df_id_col = ts_df.index.names[0]
    # convert data frame to xarray
    ts_ds = ts_df.to_xarray()
    # get only the station ids and geometries from the stations at `ts_df`
    stations_gser = stations_gdf.loc[ts_ds[stations_ts_df_id_col].values]["geometry"]
    return (
        # assign the stations geometries as indexed dimension for xvec
        ts_ds.assign_coords({stations_ts_df_id_col: stations_gser.values})
        .rename({stations_ts_df_id_col: "geometry"})
        .xvec.set_geom_indexes("geometry", crs=stations_gdf.crs)
        # add station id labels as dimensionless coordinates associated to the geometry
        .assign_coords(
            {
                stations_ts_df_id_col: (
                    "geometry",
                    stations_gser.index,
                )
            }
        )
    )


########################################################################################
# meteo utils
def get_heatwave_periods(
    ts_df: pd.DataFrame,
    *,
    heatwave_t_threshold: float | None = None,
    heatwave_n_consecutive_days: int | None = None,
    station_agg_func: AggFuncType = None,
    inter_station_agg_func: AggFuncType = None,
) -> list[tuple[dt.date, dt.date]]:
    """Get the heatwave periods from a time series of temperature measurements.

    A heatwave is defined as a period of at least `heatwave_n_consecutive_days` days
    with a temperature above `heatwave_t_threshold`.

    Parameters
    ----------
    ts_df : pd.DataFrame
        Data frame with a time series of temperature measurements at each station, in
        long or wide format.
    heatwave_t_threshold : float, optional
        The temperature threshold for a heatwave, in units of `ts_df`. If not provided,
        the value from `settings.HEATWAVE_T_THRESHOLD` is used.
    heatwave_n_consecutive_days : int, optional
        The number of consecutive days above the mean temperature threshold for the
        corresponding period to be  considered a heatwave. If not provided, the value
        from `settings.HEATWAVE_N_CONSECUTIVE_DAYS` is used.
    station_agg_func, inter_station_agg_func : str or function, optional
        How to respectively aggregate the daily temperature measurements at each station
        and the aggregated daily temperature measurements across all stations. Must be a
        string function name or a callable function, which will be passed as the `func`
        argument of `pandas.core.groupby.DataFrameGroupBy.agg`. If not provided, the
        respective values from `settings.HEATWAVE_STATION_AGG_FUNC` and
        `settings.HEATWAVE_INTER_STATION_AGG_FUNC` are used.

    Returns
    -------
    heatwave_range_df : pd.DataFrame
        Data frame with the heatwave start and end dates as columns, indexed by the
        heatwave event identifier.
    """
    # if using a multi-index, assume that it is a long-form data frame so transform it
    if isinstance(ts_df.index, pd.MultiIndex):
        # ACHTUNG: we are assuming that there is a single column with the temperature
        ts_df = long_to_wide(ts_df, variables=ts_df.columns[0])

    # process arguments
    if heatwave_t_threshold is None:
        heatwave_t_threshold = settings.HEATWAVE_T_THRESHOLD
    if heatwave_n_consecutive_days is None:
        heatwave_n_consecutive_days = settings.HEATWAVE_N_CONSECUTIVE_DAYS
    if station_agg_func is None:
        station_agg_func = settings.HEATWAVE_STATION_AGG_FUNC
    if inter_station_agg_func is None:
        inter_station_agg_func = settings.HEATWAVE_INTER_STATION_AGG_FUNC

    # find consecutive days above threshold
    # day_agg_ts_ser = getattr(
    #     getattr(ts_df.groupby(ts_df.index.date), station_agg_func)(),
    #     inter_station_agg_func,
    # )(axis="columns")
    day_agg_ts_ser = (
        ts_df.groupby(ts_df.index.date)
        .agg(station_agg_func)
        .agg(inter_station_agg_func, axis="columns")
    )
    idx = (day_agg_ts_ser >= heatwave_t_threshold).rolling(
        window=heatwave_n_consecutive_days, center=True
    ).sum() >= heatwave_n_consecutive_days

    return [
        (
            dt.datetime.combine(g.index[0], dt.time.min),
            dt.datetime.combine(g.index[-1], dt.time.max),
        )
        for _, g in idx.groupby(idx.ne(idx.shift()).cumsum())
        if g.any()
    ]


def get_heatwave_ts_df(
    ts_df: pd.DataFrame,
    *,
    heatwave_periods: list[tuple[dt.date, dt.date]] | None = None,
    heatwave_t_threshold: float | None = None,
    heatwave_n_consecutive_days: int | None = None,
    station_agg_func: str | None = None,
    inter_station_agg_func: str | None = None,
) -> pd.DataFrame:
    """Get a time series data frame for the heatwave periods.

    A heatwave is defined as a period of at least `heatwave_n_consecutive_days` days
    with a temperature above `heatwave_t_threshold`.

    Parameters
    ----------
    ts_df : pd.DataFrame
        Data frame with a time series of temperature measurements at each station, in
        long or wide format.
    heatwave_t_threshold : float, optional
        The temperature threshold for a heatwave, in units of `ts_df`. If not provided,
        the value from `settings.HEATWAVE_T_THRESHOLD` is used.
    heatwave_n_consecutive_days : int, optional
        The number of consecutive days above the mean temperature threshold for the
        corresponding period to be  considered a heatwave. If not provided, the value
        from `settings.HEATWAVE_N_CONSECUTIVE_DAYS` is used.
    station_agg_func, inter_station_agg_func : str or function, optional
        How to respectively aggregate the daily temperature measurements at each station
        and the aggregated daily temperature measurements across all stations. Must be a
        string function name or a callable function, which will be passed as the `func`
        argument of `pandas.core.groupby.DataFrameGroupBy.agg`. If not provided, the
        respective values from `settings.HEATWAVE_STATION_AGG_FUNC` and
        `settings.HEATWAVE_INTER_STATION_AGG_FUNC` are used.

    Returns
    -------
    heatwave_range_df : pd.DataFrame
        Data frame with the heatwave start and end dates as columns, indexed by the
        heatwave event identifier.
    """
    if heatwave_periods is None:
        heatwave_periods = get_heatwave_periods(
            ts_df,
            heatwave_t_threshold=heatwave_t_threshold,
            heatwave_n_consecutive_days=heatwave_n_consecutive_days,
            station_agg_func=station_agg_func,
            inter_station_agg_func=inter_station_agg_func,
        )
    if heatwave_periods:
        return (
            pd.concat(
                ts_df.loc[heatwave_start:heatwave_end].assign(
                    heatwave=f"{heatwave_start.date()}/{heatwave_end.date()}"
                )
                for heatwave_start, heatwave_end in heatwave_periods
            )
            .reset_index()
            .set_index(["heatwave", ts_df.index.name])
        )
    else:
        log(
            "No heatwave periods found, returning empty data frame.",
            level=lg.WARNING,
        )
        return pd.DataFrame(
            columns=ts_df.columns,
            index=pd.MultiIndex(
                levels=[[], []], codes=[[], []], names=["heatwave", ts_df.index.name]
            ),
        )


########################################################################################
# abstract attribute
# `DummyAttribute` and `abstract_attribute` below are hardcoded from
# github.com/rykener/better-abc to avoid relying on an unmaintained library that is not
# in conda-forge
class DummyAttribute:
    """Dummy attribute."""

    pass


def abstract_attribute(obj=None):
    """Abstract attribute."""
    if obj is None:
        obj = DummyAttribute()
    obj.__is_abstract_attribute__ = True
    return obj


########################################################################################
# logging
def ts(*, style: str = "datetime", template: str | None = None) -> str:
    """Get current timestamp as string.

    Parameters
    ----------
    style : str {"datetime", "date", "time"}
        Format the timestamp with this built-in template.
    template : str, optional
        If not None, format the timestamp with this template instead of one of the
        built-in styles.

    Returns
    -------
    ts : str
        The string timestamp.
    """
    if template is None:
        if style == "datetime":
            template = "{:%Y-%m-%d %H:%M:%S}"
        elif style == "date":
            template = "{:%Y-%m-%d}"
        elif style == "time":
            template = "{:%H:%M:%S}"
        else:  # pragma: no cover
            raise ValueError(f"unrecognized timestamp style {style!r}")

    ts = template.format(dt.datetime.now())
    return ts


def _get_logger(level: int, name: str, filename: str) -> lg.Logger:
    """Create a logger or return the current one if already instantiated.

    Parameters
    ----------
    level : int
        One of Python's logger.level constants.
    name : string
        Name of the logger.
    filename : string
        Name of the log file, without file extension.

    Returns
    -------
    logger : logging.logger
    """
    logger = lg.getLogger(name)

    # if a logger with this name is not already set up
    if not getattr(logger, "handler_set", None):
        # get today's date and construct a log filename
        log_filename = Path(settings.LOGS_FOLDER) / f"{filename}_{ts(style='date')}.log"

        # if the logs folder does not already exist, create it
        log_filename.parent.mkdir(parents=True, exist_ok=True)

        # create file handler and log formatter and set them up
        handler = lg.FileHandler(log_filename, encoding="utf-8")
        formatter = lg.Formatter("%(asctime)s %(levelname)s %(name)s %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(level)
        logger.handler_set = True

    return logger


def log(
    message: str,
    *,
    level: int | None = None,
    name: str | None = None,
    filename: str | None = None,
) -> None:
    """Write a message to the logger.

    This logs to file and/or prints to the console (terminal), depending on the current
    configuration of settings.LOG_FILE and settings.LOG_CONSOLE.

    Parameters
    ----------
    message : str
        The message to log.
    level : int, optional
        One of Python's logger.level constants. If None, the value from
        `settings.LOG_LEVEL` is used.
    name : str, optional
        Name of the logger. If None, the value from `settings.LOG_NAME` is used.
    filename : str, optional
        Name of the log file, without file extension. If None, the value from
        `settings.LOG_FILENAME` is used.
    """
    if level is None:
        level = settings.LOG_LEVEL
    if name is None:
        name = settings.LOG_NAME
    if filename is None:
        filename = settings.LOG_FILENAME

    # if logging to file is turned on
    if settings.LOG_FILE:
        # get the current logger (or create a new one, if none), then log message at
        # requested level
        logger = _get_logger(level=level, name=name, filename=filename)
        if level == lg.DEBUG:
            logger.debug(message)
        elif level == lg.INFO:
            logger.info(message)
        elif level == lg.WARNING:
            logger.warning(message)
        elif level == lg.ERROR:
            logger.error(message)

    # if logging to console (terminal window) is turned on
    if settings.LOG_CONSOLE:
        # prepend timestamp
        message = f"{ts()} {message}"

        # convert to ascii so it doesn't break windows terminals
        message = (
            unicodedata.normalize("NFKD", str(message))
            .encode("ascii", errors="replace")
            .decode()
        )

        # print explicitly to terminal in case jupyter notebook is the stdout
        if getattr(sys.stdout, "_original_stdstream_copy", None) is not None:
            # redirect captured pipe back to original
            os.dup2(sys.stdout._original_stdstream_copy, sys.__stdout__.fileno())
            sys.stdout._original_stdstream_copy = None
        with redirect_stdout(sys.__stdout__):
            print(message, file=sys.__stdout__, flush=True)



"""Quality control for CWS data.

Based on Napoly et al., 2018 (https://doi.org/10.3389/feart.2018.00118)
"""

from collections.abc import Sequence

import geopandas as gpd
import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from scipy.stats import norm
from statsmodels.robust import scale

from meteora import settings


def comparison_lineplot(
    ts_df: pd.DataFrame,
    discard_stations: Sequence,
    *,
    label_discarded: str = "discarded",
    label_kept: str = "kept",
    individual_discard_lines: bool = False,
    ax: mpl.axes.Axes = None,
):
    """Plot time series for discarded and kept stations separately.

    Parameters
    ----------
    ts_df : pandas.DataFrame
        Wide time series data frame with stations as columns and time as index.
    discard_stations : list-like
        Station ids to discard.
    label_discarded, label_kept : str, default "discarded", "kept"
        Label for discarded and kept stations, respectively.
    individual_discard_lines : bool, default False
        Plot discarded stations as individual lines (rather than line with mean and
        confidence intervals).
    ax : matplotlib.axes.Axes, default None
        Axes object to plot on. If None, a new figure is created.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axes object with the plot.
    """
    data = pd.concat(
        [
            _ts_df.reset_index()
            .melt(value_name="temperature", id_vars="time")
            .assign(label=label)
            for _ts_df, label in zip(
                [
                    ts_df[discard_stations],
                    ts_df.drop(columns=discard_stations, errors="ignore"),
                ],
                [label_discarded, label_kept],
            )
        ],
        ignore_index=True,
    )
    if ax is None:
        _, ax = plt.subplots()
    if individual_discard_lines:
        # to avoid warnings about matplotlib converters
        pd.plotting.register_matplotlib_converters()
        sns.lineplot(
            data[data["label"] == label_kept],
            x="time",
            y="temperature",
            hue="label",
            ax=ax,
        )
        data[data["label"] == label_discarded].set_index("time").rename(
            columns={"temperature": label_discarded}
        ).plot(ax=ax)
    else:
        sns.lineplot(
            data,
            x="time",
            y="temperature",
            hue="label",
            ax=ax,
        )
        ax.tick_params(axis="x", labelrotation=45)
    return ax


def get_mislocated_stations(station_gser: gpd.GeoSeries) -> list:
    """Get mislocated stations.

    When multiple stations share the same location, it is likely due to an incorrect
    set up that led to automatic location assignment based on the IP address of the
    wireless network.

    Parameters
    ----------
    station_gser : geopandas.GeoSeries
        Geoseries of station locations (points).

    Returns
    -------
    mislocated_stations : list
        List of station ids considered mislocated.
    """
    mislocated_station_ser = station_gser.duplicated(keep=False)
    return list(mislocated_station_ser[mislocated_station_ser].index)


# function to filter stations depending on the proportion of available valid
# measurements
def get_unreliable_stations(
    ts_df: pd.DataFrame, *, unreliable_threshold: float | None = None
) -> list:
    """Get stations with a high proportion of non-valid measurements.

    Parameters
    ----------
    ts_df : pandas.DataFrame
        Time series of measurements (rows) for each station (columns).
    unreliable_threshold : numeric, optional
        Proportion of non-valid measurements after which a station is considered
        unreliable. If None, the value from `settings.UNRELIABLE_THRESHOLD` is used.

    Returns
    -------
    unreliable_stations : list
        List of station ids considered unreliable.

    """
    if unreliable_threshold is None:
        unreliable_threshold = settings.UNRELIABLE_THRESHOLD

    unreliable_station_ser = (
        ts_df.isna().sum() / len(ts_df.index) > unreliable_threshold
    )
    return list(unreliable_station_ser[unreliable_station_ser].index)


def elevation_adjustment(
    ts_df: pd.DataFrame,
    station_elevation_ser: pd.Series,
    *,
    atmospheric_lapse_rate: float | None = None,
) -> pd.DataFrame:
    """Adjust temperature measurements based on station elevation.

    Parameters
    ----------
    ts_df : pandas.DataFrame
        Time series of measurements (rows) for each station (columns).
    station_elevation_ser : pandas.Series, optional
        Series of station elevations, indexed by the station id. If provided, the
        series of measurements is adjusted to account for the elevation effect.
    atmospheric_lapse_rate : numeric, optional
        Atmospheric lapse rate (in unit of `ts_df` per unit of `elevation_station_ser`)
        to account for the elevation effect. Ignored if `elevation_station_ser` is not
        provided. If None, the value from `settings.ATMOSPHERIC_LAPSE_RATE` is used.

    Returns
    -------
    adjusted_ts_df : pandas.DataFrame
        Time series of adjusted measurements (rows) for each station (columns).

    """
    if atmospheric_lapse_rate is None:
        atmospheric_lapse_rate = settings.ATMOSPHERIC_LAPSE_RATE
    station_elevation_ser = station_elevation_ser[ts_df.columns]
    return ts_df + atmospheric_lapse_rate * (
        station_elevation_ser - station_elevation_ser.mean()
    )


def get_outlier_stations(
    ts_df: pd.DataFrame,
    *,
    low_alpha: float | None = None,
    high_alpha: float | None = None,
    station_outlier_threshold: float | None = None,
) -> list:
    """Get outlier stations.

    Measurements can show suspicious deviations from a normal distribution (based on
    a modified z-score using robust Qn variance estimators). Stations with high
    proportion of such measurements can be related to radiative errors in non-shaded
    areas or other measurement errors.

    Parameters
    ----------
    ts_df : pandas.DataFrame
        Time series of measurements (rows) for each station (columns).
    low_alpha, high_alpha : numeric, optional
        Values for the lower and upper tail respectively (in proportion from 0 to 1)
        that lead to the rejection of the null hypothesis (i.e., the corresponding
        measurement does not follow a normal distribution can be considered an
        outlier). If None, the respective values from `settings.OUTLIER_LOW_ALPHA`
        and `settings.OUTLIER_HIGH_ALPHA` are used.
    station_outlier_threshold : numeric, optional
        Maximum proportion (from 0 to 1) of outlier measurements after which the
        respective station may be flagged as faulty. If None, the value from
        `settings.STATION_OUTLIER_THRESHOLD` is used.

    Returns
    -------
    outlier_stations : list
        List of station ids for stations flagged as outlier.
    """
    # def z_score(x):
    #     try:
    #         return (x - x.median()) / scale.qn_scale(x.dropna())
    #     except:
    #         print(x)
    #         raise ValueError("error")

    if low_alpha is None:
        low_alpha = settings.OUTLIER_LOW_ALPHA
    if high_alpha is None:
        high_alpha = settings.OUTLIER_HIGH_ALPHA
    if station_outlier_threshold is None:
        station_outlier_threshold = settings.STATION_OUTLIER_THRESHOLD

    low_z = norm.ppf(low_alpha)
    high_z = norm.ppf(high_alpha)
    prop_outlier_ser = ts_df.sub(ts_df.median(axis="columns"), axis="rows").div(
        ts_df.apply(scale.qn_scale, axis="columns"), axis="rows"
    ).apply(lambda z: ~z.between(low_z, high_z, inclusive="neither")).sum() / len(
        ts_df.index
    )
    outlier_station_ser = prop_outlier_ser > station_outlier_threshold
    return list(outlier_station_ser[outlier_station_ser].index)


def get_indoor_stations(
    ts_df: pd.DataFrame, *, station_indoor_corr_threshold: float | None = None
) -> list:
    """Get indoor stations.

    Stations whose time series of measurements show low correlations with the
    spatial median time series are likely set up indoors.

    Parameters
    ----------
    ts_df : pandas.DataFrame
        Time series of measurements (rows) for each station (columns).
    station_indoor_corr_threshold : numeric, optional
        Stations showing Pearson correlations (with the overall station median
        distribution) lower than this threshold are likely set up indoors. If None,
        the value from `settings.STATION_INDOOR_CORR_THRESHOLD` is used.

    Returns
    -------
    indoor_stations : list
        List of station ids for stations flagged as indoor.

    """
    if station_indoor_corr_threshold is None:
        station_indoor_corr_threshold = settings.STATION_INDOOR_CORR_THRESHOLD

    indoor_station_ser = (
        ts_df.corrwith(ts_df.median(axis="columns")) < station_indoor_corr_threshold
    )
    return list(indoor_station_ser[indoor_station_ser].index)



"""Mixins module."""

from meteora.mixins.auth import APIKeyHeaderMixin, APIKeyParamMixin
from meteora.mixins.region import RegionMixin
from meteora.mixins.stations import StationsEndpointMixin
from meteora.mixins.variables import VariablesEndpointMixin, VariablesHardcodedMixin



"""Office for Waste, Water, Energy and Air (AWEL) of the canton of Zurich."""

from collections.abc import Sequence
from datetime import date as dt_date

import pandas as pd
import pooch
import pyproj
import requests
from dateutil.relativedelta import relativedelta

from meteora import settings
from meteora.clients.base import BaseTextClient
from meteora.mixins import VariablesHardcodedMixin
from meteora.utils import DateTimeType, KwargsType, RegionType, VariablesType

# disable pooch warnings when providing `None` as "known_hash"
logger = pooch.get_logger()
logger.setLevel("WARNING")

# API endpoints
TS_ENDPOINT = (
    "https://www.web.statistik.zh.ch/awel/LoRa/data/"
    "AWEL_Sensors_LoRa_{year}{month:02}.csv"
)

# useful constants
STATIONS_GDF_ID_COL = "sensor"
TS_DF_STATIONS_ID_COL = "sensor"
TS_DF_TIME_COL = "starttime"
VARIABLES_ID_COL = "code"
VARIABLES_LABEL_COL = "description"
# see https://www.web.statistik.zh.ch/awel/LoRa/data/datenbeschrieb.txt
SENSOR_HEIGHT_COL = "magl"


# see https://www.web.statistik.zh.ch/awel/LoRa/data/datenbeschrieb.txt
VARIABLES_DICT = {
    "temperature": "measured value air temperature (in °C)",
    "humidity": "measured value relative humidity (in%)",
}
ECV_DICT = {
    "temperature": "temperature",
    "relative_humidity": "humidity",
}


class AWELClient(VariablesHardcodedMixin, BaseTextClient):
    """AWEL client (canton of Zurich).

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    sensor_height : {1, 1.8, 2, 2.5, 2.7, 3.5, 4}, default 2
        Sensor height (in m) above site location, must be a value among the following:
        1, 1.8, 2, 2.5, 2.7, 3.5, 4. The default value is 2 m.
    pooch_kwargs : dict, optional
        Keyword arguments to pass to the `pooch.retrieve` function when downloading the
        stations time series data.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # ACHTUNG: many constants are set in `GHCNH_STATIONS_COLUMNS` above
    # geom constants
    X_COL = "x"
    Y_COL = "y"
    CRS = pyproj.CRS("epsg:2056")

    # API endpoints
    _ts_endpoint = TS_ENDPOINT

    # data frame labels constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _ts_df_time_col = TS_DF_TIME_COL
    _variables_id_col = VARIABLES_ID_COL
    _variables_label_col = VARIABLES_LABEL_COL
    _variables_dict = VARIABLES_DICT
    _ecv_dict = ECV_DICT

    def __init__(
        self,
        region: RegionType,
        *,
        sensor_height: float = 2,
        pooch_kwargs: KwargsType | None = None,
        **sjoin_kwargs: KwargsType,
    ) -> None:
        """Initialize GHCN hourly client."""
        self.region = region
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs

        self._sensor_height = sensor_height

        if pooch_kwargs is None:
            pooch_kwargs = {}
        self.pooch_kwargs = pooch_kwargs

        # need to call super().__init__() to set the cache
        super().__init__()

    def _get_stations_df(self):
        today = dt_date.today()
        # since there is no stations endpoint, get the time series data from the most
        # recent month
        # while True:
        # ACHTUNG: ugly hardcoded
        max_months = 10
        for months in range(1, max_months):
            ts_df_date = today - relativedelta(months=months)
            try:
                latest_ts_df_filepath = pooch.retrieve(
                    self._ts_endpoint.format(
                        year=ts_df_date.year, month=ts_df_date.month
                    ),
                    None,
                    **self.pooch_kwargs,
                )
                break
            except requests.HTTPError:
                pass
        return (
            pd.read_csv(latest_ts_df_filepath, sep=";")
            .drop(columns=["starttime"] + list(self._variables_dict.keys()))
            .groupby("sensor")
            .first()
            .reset_index()
        )

    def _ts_params(
        self, variable_ids: Sequence, start: DateTimeType, end: DateTimeType
    ) -> dict:
        # TODO: DRY with noaa.GHCNHourlyClient?
        # process date args
        start = pd.Timestamp(start)
        end = pd.Timestamp(end)

        return dict(variable_ids=variable_ids, start=start, end=end)

    def _ts_df_from_endpoint(self, ts_params) -> pd.DataFrame:
        """Get time series data frame from endpoint."""
        # TODO: DRY with noaa.GHCNHourlyClient?
        # we override this method because we need a separate request for each station
        variable_cols = list(ts_params["variable_ids"])
        cols_to_keep = (
            [self._ts_df_stations_id_col]
            + [self._ts_df_time_col, SENSOR_HEIGHT_COL]
            + variable_cols
        )

        def _process_month_ts_df(year, month):
            try:
                month_ts_filepath = pooch.retrieve(
                    self._ts_endpoint.format(year=year, month=month),
                    None,
                    **self.pooch_kwargs,
                )
            except requests.HTTPError:
                # TODO: log?
                return pd.DataFrame()
            ts_df = pd.read_csv(month_ts_filepath, sep=";", usecols=cols_to_keep)
            # filter sensor height
            ts_df = ts_df[ts_df[SENSOR_HEIGHT_COL] == self._sensor_height]
            # filter stations
            ts_df = ts_df[
                ts_df[self._ts_df_stations_id_col].isin(self.stations_gdf.index)
            ]
            # drop (station, time) duplicates (TODO: use first and reset_index?)
            ts_df = ts_df.groupby(
                [self._ts_df_stations_id_col, self._ts_df_time_col]
            ).head(1)
            # ensure datetime
            ts_df[self._ts_df_time_col] = pd.to_datetime(ts_df[self._ts_df_time_col])
            # filter time range
            return ts_df[ts_df[self._ts_df_time_col].between(start, end)]

        start = ts_params["start"]
        end = ts_params["end"]
        date_range = pd.date_range(start=start, end=end, freq="MS")
        if len(date_range) == 0:
            date_range = [start]
        ts_df = pd.concat(
            [_process_month_ts_df(date.year, date.month) for date in date_range]
        )
        # ts_df = ts_df[ts_df[self._ts_df_time_col].between(start, end)]
        return ts_df.set_index([self._ts_df_stations_id_col, self._ts_df_time_col])

    def get_ts_df(
        self,
        variables: VariablesType,
        start: DateTimeType,
        end: DateTimeType,
    ) -> pd.DataFrame:
        """Get time series data frame.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either an AWEL variable code (string) or an
            essential climate variable (ECV) following the Meteora nomenclature
            (string).
        start, end : datetime-like, str, int, float
            Values representing the start and end of the requested data period
            respectively. Accepts any datetime-like object that can be passed to
            pandas.Timestamp.

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        return self._get_ts_df(variables, start, end)



"""Agrometeo client."""

from collections.abc import Mapping, Sequence

import pandas as pd
import pyproj

from meteora import settings
from meteora.clients.base import BaseJSONClient
from meteora.mixins import StationsEndpointMixin, VariablesEndpointMixin
from meteora.utils import CRSType, DateTimeType, KwargsType, RegionType, VariablesType

# API endpoints
BASE_URL = "https://agrometeo.ch/backend/api"
STATIONS_ENDPOINT = f"{BASE_URL}/stations"
VARIABLES_ENDPOINT = f"{BASE_URL}/sensors"
TS_ENDPOINT = f"{BASE_URL}/meteo/data"

# useful constants
LONLAT_CRS = pyproj.CRS("epsg:4326")
LV03_CRS = pyproj.CRS("epsg:21781")
# ACHTUNG: for some reason, the API mixes up the longitude and latitude columns ONLY in
# the CH1903/LV03 projection. This is why we need to swap the columns in the dict below.
GEOM_COL_DICT = {LONLAT_CRS: ["long_dec", "lat_dec"], LV03_CRS: ["lat_ch", "long_ch"]}
DEFAULT_CRS = LV03_CRS
# stations column used by the Agrometeo API (do not change)
STATIONS_GDF_ID_COL = "id"
TS_DF_STATIONS_ID_COL = "id"
TS_DF_TIME_COL = "date"
# variables name column
VARIABLES_NAME_COL = "name.en"
# variables code column
VARIABLES_ID_COL = "id"
# agrometeo sensors
# 42                       Leaf moisture III
# 43     Voltage of internal lithium battery
# 1              Temperature 2m above ground
# 4                       Relative humidity
# 6                           Precipitation
# 15              Intensity of precipitation
# 7                            Leaf moisture
# 11                         Solar radiation
# 41                           Solar Energie
# 9                          Avg. wind speed
# 14                         Max. wind speed
# 8                           Wind direction
# 22                       Temperature +10cm
# 12                    Luxmeter after Lufft
# 10                                ETP-Turc
# 24                              ETo-PenMon
# 13                               Dew point
# 18                       Real air pressure
# 2                    Soil temperature +5cm
# 19                  Soil temperature -20cm
# 3                   Soil temperature -10cm
# 5                       Soil moisture -5cm
# 20                   Pressure on sea level
# 17                        Leaf moisture II
# 25                     Soil moisture -30cm
# 26                     Soil moisture -50cm
# 39                                  unused
# 33                 Temperature in leafzone
# 32                         battery voltage
# 21                         min. wind speed
# 23                        Temperatur +20cm
# 27                  Temperatur in Pflanze1
# 28                  Temperatur in Pflanze1
# 29                                    UVAB
# 30                                     UVA
# 31                                     UAB
# 34                Air humidity in leafzone
# 35             Photosyth. active radiation
# 36                  Soil temperature -10cm
# 37                Temperatur 2m unbelüftet
# 38           elative Luftfeuchtigkeit +5cm
# 40                     Precip. Radolan Day
# 100                                   Hour
# 101                                   Year
# 102                            Day of year
# 103                           Degree hours
# 104                 Density of sporulation
# 105                           Leaf surface
ECV_DICT = {
    # precipitation
    "precipitation": 6,  # "Precipitation"
    # pressure
    "pressure": 18,  # "Real air pressure"
    # radiation budget
    "radiation_shortwave": 11,  # "Solar radiation"
    # temperature
    "temperature": 1,  # "Temperature 2m above ground"
    # water vapour
    "dew_point": 13,  # "Dew point"
    "relative_humidity": 4,  # "Relative humidity"
    # wind
    "wind_speed": 9,  # "Avg. wind speed"
    "wind_direction": 8,  # "Wind direction"
}
API_DT_FMT = "%Y-%m-%d"
SCALE = "none"
MEASUREMENT = "avg"


class AgrometeoClient(StationsEndpointMixin, VariablesEndpointMixin, BaseJSONClient):
    """Agrometeo client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    crs : str, dict or pyproj.CRS, optional
        The coordinate reference system (CRS) to be used. For Agrometeo, the
        provided value must be equivalent to either the EPSG:21781 (default) or
        EPSG:4326.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # API endpoints
    _stations_endpoint = STATIONS_ENDPOINT
    _ts_endpoint = TS_ENDPOINT
    _variables_endpoint = VARIABLES_ENDPOINT

    # data frame labels constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _ts_df_time_col = TS_DF_TIME_COL
    _variables_id_col = VARIABLES_ID_COL
    # _variables_name_col = VARIABLES_NAME_COL
    _ecv_dict = ECV_DICT

    def __init__(
        self,
        region: RegionType,
        *,
        crs: CRSType | None = None,
        **sjoin_kwargs: KwargsType,
    ) -> None:
        """Initialize Agrometeo client."""
        # ACHTUNG: CRS must be either EPSG:4326 or EPSG:21781
        # ACHTUNG: CRS must be set before region
        if crs is not None:
            crs = pyproj.CRS(crs)
        else:
            crs = DEFAULT_CRS
        self.CRS = crs
        # self._variables_name_col = variables_name_col or VARIABLES_NAME_COL
        try:
            self.X_COL, self.Y_COL = GEOM_COL_DICT[self.CRS]
        except KeyError:
            raise ValueError(
                f"CRS must be among {list(GEOM_COL_DICT.keys())}, got {self.CRS}"
            )

        self.region = region
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs

        # need to call super().__init__() to set the cache
        super().__init__()

    def _stations_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        return pd.DataFrame(response_content["data"])

    def _variables_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        variables_df = pd.json_normalize(response_content["data"])
        # ACHTUNG: need to strip strings, at least in variables name column. Note
        # that *it seems* that the integer type of variable code column is inferred
        # correctly
        variables_df[VARIABLES_NAME_COL] = variables_df[VARIABLES_NAME_COL].str.strip()
        return variables_df

    def _ts_params(
        self,
        variable_ids: Sequence,
        start: DateTimeType,
        end: DateTimeType,
        scale: str | None = None,
        measurement: str | None = None,
    ) -> dict:
        # process date args
        start_date = pd.Timestamp(start).strftime(API_DT_FMT)
        end_date = pd.Timestamp(end).strftime(API_DT_FMT)
        # process scale and measurement args
        if scale is None:
            # the API needs it to be lowercase
            scale = SCALE
        if measurement is None:
            measurement = MEASUREMENT

        _stations_ids = self.stations_gdf.index.astype(str)

        return {
            "from": start_date,
            "to": end_date,
            "scale": scale,
            "sensors": ",".join(
                f"{variable_id}:{measurement}" for variable_id in variable_ids
            ),
            "stations": ",".join(_stations_ids),
        }

    def _ts_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        # parse the response as a data frame
        ts_df = pd.json_normalize(response_content["data"]).set_index(
            self._ts_df_time_col
        )
        ts_df.index = pd.to_datetime(ts_df.index)
        ts_df.index.name = self._ts_df_time_col

        # ts_df.columns = self.stations_gdf[STATIONS_ID_COL]
        # ACHTUNG: note that agrometeo returns the data indexed by keys of the form
        # "{station_id}_{variable_code}_{measurement}". We can ignore the latter and
        # convert to a two-level (station, variable) multi index
        ts_df.columns = (
            ts_df.columns.str.split("_")
            .str[:-1]
            .map(tuple)
            .rename([self._ts_df_stations_id_col, "variable"])
        )
        # convert station and variable ids to integer
        # ts_df.columns = ts_df.columns.set_levels(
        #     ts_df.columns.levels["station"].astype(int), level="station"
        # )
        for level_i, level_name in enumerate(ts_df.columns.names):
            ts_df.columns = ts_df.columns.set_levels(
                ts_df.columns.levels[level_i].astype(int), level=level_name
            )

        # convert to long form and return it
        return ts_df.stack(
            level=self._ts_df_stations_id_col, future_stack=True
        ).swaplevel()

    def get_ts_df(
        self,
        variables: VariablesType,
        start: DateTimeType,
        end: DateTimeType,
        *,
        scale: str | None = None,
        measurement: str | None = None,
    ) -> pd.DataFrame:
        """Get time series data frame.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either an Agrometeo variable code (integer or
            string) or an essential climate variable (ECV) following the Meteora
            nomenclature (string).
        start, end : datetime-like, str, int, float
            Values representing the start and end of the requested data period
            respectively. Accepts any datetime-like object that can be passed to
            pandas.Timestamp.
        scale : {"hour", "day", "month", "year"}, optional
            Temporal scale of the measurements. If None, returns the finest scale, i.e.,
            10 minutes.
        measurement : {"min", "avg", "max"}, optional
            Whether the measurement values correspond to the minimum, average or maximum
            value for the required temporal scale. If None, returns the average. Ignored
            if `scale` is None.

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        ts_df = self._get_ts_df(
            variables, start, end, scale=scale, measurement=measurement
        )
        # filter time range, otherwise, for some reason, agrometeo API includes one day
        # after
        # TODO: dry with Meteocat, perhaps a global approach in the base client
        time_ser = ts_df.index.get_level_values(settings.TIME_COL).to_series()
        tz = time_ser.dt.tz
        return ts_df.loc[
            (
                slice(None),
                time_ser.between(
                    pd.Timestamp(start, tz=tz),
                    pd.Timestamp(end, tz=tz),
                    inclusive="both",
                ),
            ),
            :,
        ]



"""Region mixins."""

import geopandas as gpd
from pyogrio.errors import DataSourceError
from shapely import geometry
from shapely.geometry.base import BaseGeometry

from meteora.utils import CRSType, KwargsType, RegionType

try:
    import osmnx as ox
except ImportError:
    ox = None


class RegionMixin:
    """Mixin class to add a `region` attribute to a class."""

    @property
    def region(self) -> gpd.GeoDataFrame | None:
        """The region as a GeoDataFrame."""
        return self._region

    @region.setter
    def region(
        self,
        region: RegionType,
    ):
        self._region = self._process_region_arg(region, crs=getattr(self, "CRS", None))

    def _process_region_arg(
        self,
        region: RegionType,
        *,
        crs: CRSType | None = None,
        **geocode_to_gdf_kwargs: KwargsType,
    ) -> gpd.GeoDataFrame | None:
        """Process the region argument.

        Parameters
        ----------
        region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
            The region to process. This can be either:
            -  A string with a place name (Nominatim query) to geocode.
            -  A sequence with the west, south, east and north bounds.
            -  A geometric object, e.g., shapely geometry, or a sequence of geometric
               objects. In such a case, the value is passed as the `data` argument of
               the GeoSeries constructor, and needs to be in the same CRS as the one
               used by the client's class (i.e., the `CRS` class attribute).
            -  A geopandas geo-series or geo-data frame.
            -  A filename or URL, a file-like object opened in binary ('rb') mode, or a
               Path object that will be passed to `geopandas.read_file`.
        crs : str, dict or pyproj.CRS, optional
            The coordinate reference system (CRS) of the provided region. It can be any
            CRS-like object accepted by geopandas. Ignored if `region` is a string
            corresponding to a place name, a geopandas geo-series or geo-data frame with
            its CRS attribute set or a filename, URL or file-like object.
        geocode_to_gdf_kwargs : dict, optional
            Keyword arguments to pass to `geocode_to_gdf` if `region` is a string
            corresponding to a place name (Nominatim query).

        Returns
        -------
        gdf : GeoDataFrame
            The processed region as a GeoDataFrame, in the CRS used by the client's
            class. A value of None is returned when passing a place name (Nominatim
            query) but osmnx is not installed.
        """
        # crs : Any, optional
        # Coordinate Reference System of the provided `region`. Ignored if `region` is a
        # string corresponding to a place name, a geopandas geo-series or geo-data frame
        # with its CRS attribute set or a filename, URL or file-like object. Can be
        # anything accepted by `pyproj.CRS.from_user_input()`, such as an authority
        # string (eg “EPSG:4326”) or a WKT string.

        if not isinstance(region, gpd.GeoDataFrame):
            # naive geometries
            if not isinstance(region, gpd.GeoSeries) and (
                hasattr(region, "__iter__")
                and not isinstance(region, str)
                or isinstance(region, BaseGeometry)
            ):
                # if region is a sequence (other than a string)
                # use the hasattr to avoid AttributeError when region is a BaseGeometry
                if hasattr(region, "__len__"):
                    if len(region) == 4 and isinstance(region[0], (int, float)):
                        # if region is a sequence of 4 numbers, assume it's a bounding
                        # box
                        region = geometry.box(*region)
                # otherwise, assume it's a geometry or sequence of geometries that can
                # be passed as the `data` argument of the GeoSeries constructor
                region = gpd.GeoSeries(region, crs=self.CRS)
            if isinstance(region, gpd.GeoSeries):
                # if we have a GeoSeries, convert it to a GeoDataFrame so that we can
                # use the same code
                region = gpd.GeoDataFrame(
                    geometry=region, crs=getattr(region, "crs", self.CRS)
                )
            else:
                # at this point, we assume that this is either file-like or a Nominatim
                # query
                try:
                    region = gpd.read_file(region)
                except (DataSourceError, AttributeError):
                    #             if ox is None:
                    #                 lg.warning(
                    #                     """
                    # Using a Nominatim query as `region` argument requires osmnx.
                    # You can install it using conda or pip.
                    # """
                    #                 )
                    #                 return

                    if geocode_to_gdf_kwargs is None:
                        geocode_to_gdf_kwargs = {}
                    region = ox.geocode_to_gdf(region, **geocode_to_gdf_kwargs).iloc[:1]

        return region.to_crs(self.CRS)



"""AEMET client."""

from collections.abc import Mapping

import pandas as pd
import pyproj

from meteora import settings, utils
from meteora.clients.base import BaseJSONClient
from meteora.mixins import (
    APIKeyParamMixin,
    StationsEndpointMixin,
    VariablesEndpointMixin,
)
from meteora.utils import KwargsType, RegionType, VariablesType

# API endpoints
BASE_URL = "https://opendata.aemet.es/opendata/api"
STATIONS_ENDPOINT = (
    f"{BASE_URL}/valores/climatologicos/inventarioestaciones/todasestaciones"
)
VARIABLES_ENDPOINT = TS_ENDPOINT = f"{BASE_URL}/observacion/convencional/todas"

# useful constants
# ACHTUNG: in Aemet, the station id col is "indicativo" in the stations endpoint but
# "idema" in the data endpoint
STATIONS_GDF_ID_COL = "indicativo"
TS_DF_STATIONS_ID_COL = "idema"
TS_DF_TIME_COL = "fint"
VARIABLES_ID_COL = "id"
ECV_DICT = {
    # precipitation
    "precipitation": "prec",
    # pressure
    "pressure": "pres",
    # temperature
    "temperature": "ta",
    # wind speed and direction
    "wind_speed": "vv",
    "wind_direction": "dv",
    # water vapour
    "dew_point": "tpr",
    "relative_humidity": "hr",
}


class AemetClient(
    APIKeyParamMixin,
    StationsEndpointMixin,
    VariablesEndpointMixin,
    BaseJSONClient,
):
    """AEMET client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    api_key : str
        AEMET API key.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # geom constants
    X_COL = "longitud"
    Y_COL = "latitud"
    CRS = pyproj.CRS("epsg:4326")

    # API endpoints
    _stations_endpoint = STATIONS_ENDPOINT
    _ts_endpoint = TS_ENDPOINT
    _variables_endpoint = VARIABLES_ENDPOINT

    # data frame labels constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _ts_df_time_col = TS_DF_TIME_COL
    # _variables_name_col = VARIABLES_NAME_COL
    _variables_id_col = VARIABLES_ID_COL
    _ecv_dict = ECV_DICT

    # auth constants
    _api_key_param_name = "api_key"
    # request_headers = {"cache-control": "no-cache"}

    def __init__(
        self, region: RegionType, api_key: str, **sjoin_kwargs: KwargsType
    ) -> None:
        """Initialize AEMET client."""
        self.region = region
        self._api_key = api_key
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs
        # need to call super().__init__() to set the cache
        super().__init__()

    @property
    def request_headers(self) -> dict:
        """Request headers."""
        try:
            return self._request_headers
        except AttributeError:
            self._request_headers = super().request_headers | {
                "cache-control": "no-cache"
            }
        return self._request_headers

    def _stations_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        # response_content returns a dict with urls, where the one under the "datos" key
        # contains the JSON data
        stations_df = pd.read_json(response_content["datos"], encoding="latin1")
        for col in [self.X_COL, self.Y_COL]:
            stations_df[col] = utils.dms_to_decimal(stations_df[col])
        return stations_df

    def _variables_df_from_content(self, response_json) -> pd.DataFrame:
        return pd.json_normalize(
            pd.read_json(response_json["metadatos"], encoding="latin1")["campos"]
        )

    @property
    def variables_df(self) -> pd.DataFrame:
        """Variables dataframe."""
        try:
            return self._variables_df
        except AttributeError:
            with self._session.cache_disabled():
                response_content = self._get_content_from_url(self._variables_endpoint)
            self._variables_df = self._variables_df_from_content(response_content)
            return self._variables_df

    def _ts_df_from_content(self, response_content: Mapping) -> pd.DataFrame:
        # response_content returns a dict with urls, where the one under the "datos" key
        # contains the JSON data
        ts_df = pd.read_json(response_content["datos"], encoding="latin1")
        # filter only stations from the region
        return ts_df[
            ts_df[self._ts_df_stations_id_col].isin(self.stations_gdf.index)
        ].set_index([self._ts_df_stations_id_col, self._ts_df_time_col])

    def get_ts_df(
        self,
        variables: VariablesType,
    ) -> pd.DataFrame:
        """Get time series data frame for the last 24h.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either an AEMET variable code (integer or
            string) or an essential climate variable (ECV) following the Meteora
            nomenclature (string).

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        # disable cache since the endpoint returns the latest 24h of data
        with self._session.cache_disabled():
            return self._get_ts_df(variables)



"""Iowa Environmental Mesonet (IEM) client."""

import abc
import io
from collections.abc import Sequence

import geopandas as gpd
import pandas as pd
import pyproj

from meteora import settings
from meteora.clients.base import BaseTextClient
from meteora.mixins import StationsEndpointMixin, VariablesHardcodedMixin
from meteora.utils import DateTimeType, KwargsType, RegionType, VariablesType

# API endpoints
BASE_URL = "https://mesonet.agron.iastate.edu"
# STATIONS_ENDPOINT = (
#     f"{BASE_URL}/sites/networks.php?special=allasos&format=csv&nohtml=on"
# )

# useful constants
STATIONS_GDF_ID_COL = "id"
TS_DF_STATIONS_ID_COL = "station"
VARIABLES_ID_COL = "code"
VARIABLES_LABEL_COL = "description"

# ASOS 1 minute https://mesonet.agron.iastate.edu/cgi-bin/request/asos1min.py?help
ONEMIN_STATIONS_ENDPOINT = f"{BASE_URL}/geojson/network/ASOS1MIN.geojson?only_online=0"
ONEMIN_TS_ENDPOINT = f"{BASE_URL}/cgi-bin/request/asos1min.py"
# tmpf: Air Temperature [F]
# dwpf: Dew Point Temperature [F]
# sknt: Wind Speed [knots]
# drct: Wind Direction
# gust_drct: 5 sec gust Wind Direction
# gust_sknt: 5 sec gust Wind Speed [knots]
# vis1_coeff: Visibility 1 Coefficient
# vis1_nd: Visibility 1 Night/Day
# vis2_coeff: Visibility 2 Coefficient
# vis2_nd: Visibility 2 Night/Day
# vis3_coeff: Visibility 3 Coefficient
# vis3_nd: Visibility 3 Night/Day
# ptype: Precip Type Code
# precip: 1 minute precip [inches]
# pres1: Sensor 1 Station Pressure [inches]
# pres2: Sensor 2 Station Pressure [inches]
# pres3: Sensor 3 Station Pressure [inches]
ONEMIN_VARIABLES_DICT = {
    "tmpf": "Air Temperature",
    "dwpf": "Dew Point Temperature",
    "sknt": "Wind Speed",
    "drct": "Wind Direction",
    "pres1": "Sensor 1 Station Pressure",
    "precip": "1 minute precip",
}
ONEMIN_ECV_DICT = {
    # precipitation
    "precipitation": "precip",
    # pressure
    "pressure": "pres1",
    # temperature
    "temperature": "tmpf",
    # water vapour
    "dew_point": "dwpf",
    # wind
    "wind_speed": "sknt",
    "wind_direction": "drct",
}
ONEMIN_TS_DF_TIME_COL = "valid(UTC)"

# METAR/ASOS https://mesonet.agron.iastate.edu/cgi-bin/request/asos.py?help
METAR_STATIONS_ENDPOINT = f"{BASE_URL}/geojson/network/AZOS.geojson"
METAR_TS_ENDPOINT = f"{BASE_URL}/cgi-bin/request/asos.py"
# see https://www.weather.gov/media/asos/aum-toc.pdf
# station: three or four character site identifier
# valid: timestamp of the observation
# tmpf: Air Temperature in Fahrenheit, typically @ 2 meters
# dwpf: Dew Point Temperature in Fahrenheit, typically @ 2 meters
# relh: Relative Humidity in %
# drct: Wind Direction in degrees from *true* north
# sknt: Wind Speed in knots
# p01i: One hour precipitation for the period from the observation time to the time of
#       the previous hourly precipitation reset. This varies slightly by site. Values
#       are in inches. This value may or may not contain frozen precipitation melted by
#       some device on the sensor or estimated by some other means. Unfortunately, we do
#       not know of an authoritative database denoting which station has which sensor.
# alti: Pressure altimeter in inches
# mslp: Sea Level Pressure in millibar
# vsby: Visibility in miles
# gust: Wind Gust in knots
# skyc1: Sky Level 1 Coverage
# skyc2: Sky Level 2 Coverage
# skyc3: Sky Level 3 Coverage
# skyc4: Sky Level 4 Coverage
# skyl1: Sky Level 1 Altitude in feet
# skyl2: Sky Level 2 Altitude in feet
# skyl3: Sky Level 3 Altitude in feet
# skyl4: Sky Level 4 Altitude in feet
# wxcodes: Present Weather Codes (space separated)
# feel: Apparent Temperature (Wind Chill or Heat Index) in Fahrenheit
# ice_accretion_1hr: Ice Accretion over 1 Hour (inches)
# ice_accretion_3hr: Ice Accretion over 3 Hours (inches)
# ice_accretion_6hr: Ice Accretion over 6 Hours (inches)
# peak_wind_gust: Peak Wind Gust (from PK WND METAR remark) (knots)
# peak_wind_drct: Peak Wind Gust Direction (from PK WND METAR remark) (deg)
# peak_wind_time: Peak Wind Gust Time (from PK WND METAR remark)
# metar: unprocessed reported observation in METAR format
METAR_VARIABLES_DICT = {
    "tmpf": "Air Temperature",
    "dwpf": "Dew Point Temperature",
    "relh": "Relative Humidity",
    "sknt": "Wind Speed",
    "drct": "Wind Direction",
    "mslp": "Sea Level Pressure in millibar",
    "p01i": "1 minute precip",
}
METAR_ECV_DICT = {
    # precipitation
    "precipitation": "p01i",
    # pressure
    "pressure": "mslp",
    # temperature
    "temperature": "tmpf",
    # water vapour
    "dew_point": "dwpf",
    "relative_humidity": "relh",
    # wind
    "wind_speed": "sknt",
    "wind_direction": "drct",
}
METAR_TS_DF_TIME_COL = "valid"


class IEMClient(
    StationsEndpointMixin, VariablesHardcodedMixin, BaseTextClient, abc.ABC
):
    """Abstract Iowa Environmental Mesonet (IEM) client."""

    # geom constants
    CRS = pyproj.CRS("epsg:4326")

    # data frame label constants
    _stations_gdf_id_col = STATIONS_GDF_ID_COL
    _ts_df_stations_id_col = TS_DF_STATIONS_ID_COL
    _variables_id_col = VARIABLES_ID_COL
    _variables_label_col = VARIABLES_LABEL_COL

    def __init__(self, region: RegionType, **sjoin_kwargs: KwargsType) -> None:
        """Initialize Iowa Environmental Mesonet (IEM) client."""
        self.region = region
        if sjoin_kwargs is None:
            sjoin_kwargs = settings.SJOIN_KWARGS.copy()
        self.SJOIN_KWARGS = sjoin_kwargs

        # need to call super().__init__() to set the cache
        super().__init__()

    def _get_stations_gdf(self) -> gpd.GeoDataFrame:
        """Get a GeoDataFrame featuring the stations data for the given region.

        Returns
        -------
        stations_gdf : gpd.GeoDataFrame
            The stations data for the given region as a GeoDataFrame.

        """
        # ACHTUNG: here we "bypass" `self._get_stations_df` because the stations are
        # provided as GeoJSON
        stations_gdf = gpd.read_file(self._stations_endpoint)
        # filter the stations
        # TODO: do we need to copy the dict to avoid reference issues?
        _sjoin_kwargs = self.SJOIN_KWARGS.copy()
        # predicate = _sjoin_kws.pop("predicate", SJOIN_PREDICATE)
        return stations_gdf.sjoin(self.region[["geometry"]], **_sjoin_kwargs)[
            stations_gdf.columns
        ]

    def _ts_params(
        self, variable_ids: Sequence, start: DateTimeType, end: DateTimeType
    ) -> dict:
        # process date args
        start = pd.Timestamp(start)
        end = pd.Timestamp(end)

        return {
            "year1": start.year,
            "month1": start.month,
            "day1": start.day,
            "year2": end.year,
            "month2": end.month,
            "day2": end.day,
            self._vars_param: ",".join(variable_ids),
            "station": ",".join(self.stations_gdf.index),
        }

    def _ts_df_from_content(self, response_content: io.StringIO) -> pd.DataFrame:
        ts_df = pd.read_csv(
            response_content,
            na_values="M",
        )
        return (
            ts_df.assign(
                **{self._ts_df_time_col: pd.to_datetime(ts_df[self._ts_df_time_col])}
            )
            .groupby(["station", self._ts_df_time_col])
            .first(skipna=True)
        )

    def _post_process_ts_df(self, ts_df: pd.DataFrame) -> pd.DataFrame:
        # In this case:
        # - avoid sorting on index as data is already sorted
        # - avoid to_numeric as data is already numeric
        return ts_df

    def get_ts_df(
        self,
        variables: VariablesType,
        start: DateTimeType,
        end: DateTimeType,
    ) -> pd.DataFrame:
        """Get time series data frame for a given station.

        Parameters
        ----------
        variables : str, int or list-like of str or int
            Target variables, which can be either an IEM variable code (integer or
            string) or an essential climate variable (ECV) following the Meteora
            nomenclature (string).
        start, end : datetime-like, str, int, float
            Values representing the start and end of the requested data period
            respectively. Accepts any datetime-like object that can be passed to
            pandas.Timestamp.

        Returns
        -------
        ts_df : pandas.DataFrame
            Long form data frame with a time series of measurements (second-level index)
            at each station (first-level index) for each variable (column).
        """
        return self._get_ts_df(variables, start, end)


class ASOSOneMinIEMClient(IEMClient):
    """ASOS 1 minute Iowa Environmental Mesonet (IEM) client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # API endpoints
    _stations_endpoint = ONEMIN_STATIONS_ENDPOINT
    _ts_endpoint = ONEMIN_TS_ENDPOINT

    # data frame labels constants
    _ts_df_time_col = ONEMIN_TS_DF_TIME_COL
    _variables_dict = ONEMIN_VARIABLES_DICT
    _ecv_dict = ONEMIN_ECV_DICT
    _vars_param = "vars"


class METARASOSIEMClient(IEMClient):
    """METAR/ASOS Iowa Environmental Mesonet (IEM) client.

    Parameters
    ----------
    region : str, Sequence, GeoSeries, GeoDataFrame, PathLike, or IO
        The region to process. This can be either:

        -  A string with a place name (Nominatim query) to geocode.
        -  A sequence with the west, south, east and north bounds.
        -  A geometric object, e.g., shapely geometry, or a sequence of geometric
           objects. In such a case, the value will be passed as the `data` argument of
           the GeoSeries constructor, and needs to be in the same CRS as the one used by
           the client's class (i.e., the `CRS` class attribute).
        -  A geopandas geo-series or geo-data frame.
        -  A filename or URL, a file-like object opened in binary ('rb') mode, or a Path
           object that will be passed to `geopandas.read_file`.
    sjoin_kwargs : dict, optional
        Keyword arguments to pass to the `geopandas.sjoin` function when filtering the
        stations within the region. If None, the value from `settings.SJOIN_KWARGS` is
        used.
    """

    # API endpoints
    _stations_endpoint = METAR_STATIONS_ENDPOINT
    _ts_endpoint = METAR_TS_ENDPOINT

    # data frame labels constants
    _ts_df_time_col = METAR_TS_DF_TIME_COL
    _variables_dict = METAR_VARIABLES_DICT
    _ecv_dict = METAR_ECV_DICT
    _vars_param = "data"



"""Base abstract classes for meteo station datasets."""

import abc
import io
import logging as lg
import re
import time
from collections.abc import Mapping

import geopandas as gpd
import pandas as pd
import pyproj
import requests
import requests_cache

from meteora import settings, utils
from meteora.mixins import RegionMixin
from meteora.utils import KwargsType, VariablesType

__all__ = [
    "BaseJSONClient",
    "BaseTextClient",
]


class BaseClient(RegionMixin, abc.ABC):
    """Meteora base client."""

    def __init__(self, *args, **kwargs):
        # if use_cache is None:
        #     use_cache = settings.USE_CACHE
        if settings.USE_CACHE:  # if use_cache:
            session = requests_cache.CachedSession(
                cache_name=settings.CACHE_NAME,
                backend=settings.CACHE_BACKEND,
                expire_after=settings.CACHE_EXPIRE,
            )
        else:
            session = requests.Session()
        self._session = session

    @utils.abstract_attribute
    def X_COL(self) -> str:  # pylint: disable=invalid-name
        """Name of the column with longitude coordinates."""
        pass

    @utils.abstract_attribute
    def Y_COL(self) -> str:  # pylint: disable=invalid-name
        """Name of the column with latitude coordinates."""
        pass

    @utils.abstract_attribute
    def CRS(self) -> pyproj.CRS:  # pylint: disable=invalid-name
        """CRS of the data source."""
        pass

    @utils.abstract_attribute
    def _stations_gdf_id_col(self) -> str:
        """Column with the station IDs in the `stations_gdf` returned by the API."""
        pass

    @utils.abstract_attribute
    def _ts_df_time_col(self) -> str:
        """Column with the timestamps in the `ts_df` returned by the API."""
        pass

    @utils.abstract_attribute
    def _ts_df_stations_id_col(self) -> str:
        """Column with the station IDs in the `ts_df` returned by the API."""
        pass

    @property
    def request_headers(self) -> dict:
        """Request headers."""
        return {}

    @property
    def request_params(self) -> dict:
        """Request parameters."""
        return {}

    def _get(
        self,
        url: str,
        *,
        params: KwargsType = None,
        headers: KwargsType = None,
        **request_kwargs: KwargsType,
    ) -> requests.Response:
        """Get response for the url (from the cache or from the API).

        Parameters
        ----------
        url : str
            URL to request.
        params : dict, optional
            Parameters to pass to the request. They will be added to the default params
            set in the `request_params` property.
        headers : dict, optional
            Headers to pass to the request. They will be added to the default headers
            set in the `request_headers` property.
        request_kwargs : dict, optional
            Additional keyword arguments to pass to `requests.get`. If None, the value
            from `settings.REQUEST_KWARGS` is used.

        Returns
        -------
        response : requests.Response
            Response object from the server.
        """
        _params = self.request_params.copy()
        _headers = self.request_headers.copy()
        _request_kwargs = settings.REQUEST_KWARGS.copy()
        if params is not None:
            _params.update(params)
        if headers is not None:
            _headers.update(headers)
        if request_kwargs is not None:
            _request_kwargs.update(request_kwargs)

        return self._session.get(
            url, params=_params, headers=_headers, **_request_kwargs
        )

    @abc.abstractmethod
    def _get_content_from_response(self, response: requests.Response):
        pass

    def _get_content_from_url(
        self,
        url: str,
        params: KwargsType = None,
        headers: KwargsType = None,
        request_kwargs: KwargsType = None,
        pause: int | None = None,
        error_pause: int | None = None,
    ):
        """Get the response content from a given URL.

        Parameters
        ----------
        url : str
            URL to request.
        params : dict, optional
            Parameters to pass to the request. They will be added to the default params
            set in the `request_params` property.
        headers : dict, optional
            Headers to pass to the request. They will be added to the default headers
            set in the `request_headers` property.
        request_kwargs : dict, optional
            Additional keyword arguments to pass to `requests.get`. If None, the value
            from `settings.REQUEST_KWARGS` is used.
        pause : int, optional
            How long to pause before request, in seconds. If None, the value from
            `settings.PAUSE` is used.
        error_pause : int, optional
            How long to pause in seconds before re-trying request if error. If None, the
            value from `settings.ERROR_PAUSE` is used.

        Returns
        -------
        response_content
            Response content.
        """
        if request_kwargs is None:
            request_kwargs = {}
        response = self._get(url, params=params, headers=headers, **request_kwargs)
        sc = response.status_code
        try:
            response_content = self._get_content_from_response(response)
        except Exception:  # pragma: no cover
            domain = re.findall(r"(?s)//(.*?)/", url)[0]
            if sc in {429, 504}:
                # 429 is 'too many requests' and 504 is 'gateway timeout' from
                # server overload: handle these by pausing then recursively
                # re-trying until we get a valid response from the server
                if error_pause is None:
                    error_pause = settings.ERROR_PAUSE
                utils.log(
                    f"{domain} returned {sc}: retry in {error_pause} secs",
                    level=lg.WARNING,
                )
                time.sleep(error_pause)
                # note that this is a recursive call
                response_content = self._get_content_from_url(
                    url,
                    params=params,
                    headers=headers,
                    request_kwargs=request_kwargs,
                    pause=pause,
                    error_pause=error_pause,
                )
            else:
                # else, this was an unhandled status code, throw an exception
                utils.log(f"{domain} returned {sc}", level=lg.ERROR)
                raise Exception(
                    f"Server returned:\n{response} {response.reason}\n{response.text}"
                )

        return response_content

    # stations
    def _get_stations_gdf(self) -> gpd.GeoDataFrame:
        """Get a GeoDataFrame featuring the stations data for the given region.

        Returns
        -------
        stations_gdf : gpd.GeoDataFrame
            The stations data for the given region as a GeoDataFrame.

        """
        stations_df = self._get_stations_df()
        stations_gdf = gpd.GeoDataFrame(
            stations_df,
            geometry=gpd.points_from_xy(
                stations_df[self.X_COL], stations_df[self.Y_COL]
            ),
            crs=self.CRS,
        )
        # filter the stations
        # TODO: do we need to copy the dict to avoid reference issues?
        _sjoin_kwargs = self.SJOIN_KWARGS.copy()
        # predicate = _sjoin_kws.pop("predicate", SJOIN_PREDICATE)
        return stations_gdf.sjoin(self.region[["geometry"]], **_sjoin_kwargs)[
            stations_gdf.columns
        ]

    @property
    def stations_gdf(self) -> gpd.GeoDataFrame:
        """Geo-data frame with stations data."""
        try:
            return self._stations_gdf
        except AttributeError:
            # rename here because some clients may override `_get_stations_gdf`
            self._stations_gdf = (
                self._get_stations_gdf()
                .rename(columns={self._stations_gdf_id_col: settings.STATIONS_ID_COL})
                .set_index(settings.STATIONS_ID_COL)
            )
            return self._stations_gdf

    # time series data
    @utils.abstract_attribute
    def _ts_endpoint(self) -> str:
        pass

    def _ts_params(self, variable_ids, *args, **kwargs) -> dict:
        return {"variable_ids": variable_ids, **kwargs}

    def _post_process_ts_df(self, ts_df: pd.DataFrame) -> pd.DataFrame:
        return ts_df.apply(pd.to_numeric, axis="columns").sort_index()

    def _rename_variables_cols(
        self, ts_df: pd.DataFrame, variable_id_ser: pd.Series
    ) -> pd.DataFrame:
        # TODO: avoid this if the user provided variable codes (in which case the dict
        # maps variable codes to variable codes)?
        # also keep only columns of requested variables
        return ts_df[variable_id_ser].rename(
            columns={
                variable_id: variable
                for variable, variable_id in variable_id_ser.items()
            }
        )

    def _ts_df_from_endpoint(self, ts_params: Mapping) -> pd.DataFrame:
        # perform request
        response_content = self._get_content_from_url(
            self._ts_endpoint, params=ts_params
        )

        # process response content into a time series data frame
        return self._ts_df_from_content(response_content)

    def _get_ts_df(self, variables: VariablesType, *args, **kwargs) -> pd.DataFrame:
        # process the variables arg
        variable_id_ser = self._get_variable_id_ser(variables)

        # prepare base request parameters
        ts_params = self._ts_params(variable_id_ser, *args, **kwargs)

        # perform request
        ts_df = self._ts_df_from_endpoint(ts_params)

        # ACHTUNG: do NOT set the station, time multi-index here because this is already
        # done in `_ts_df_from_content` in many cases since it results from groupby,
        # stack or pivot operations
        # # set station, time multi-index
        # ts_df = ts_df.set_index([self._stations_id_col, self._time_col])

        # ensure that we return the variable column names as provided by the user in the
        # `variables` argument (e.g., if the user provided variable codes, use
        # variable codes in the column names).
        ts_df = self._rename_variables_cols(ts_df, variable_id_ser)

        # apply a generic post-processing function (by default, ensuring numeric dtypes
        # and sorting)
        ts_df = self._post_process_ts_df(ts_df)

        # rename stations and id labels in multi-level index
        return ts_df.rename_axis(
            index={
                self._ts_df_stations_id_col: settings.STATIONS_ID_COL,
                self._ts_df_time_col: settings.TIME_COL,
            }
        )


class BaseJSONClient(BaseClient):
    """Base class for JSON clients."""

    def _get_content_from_response(self, response: requests.Response) -> dict:
        return response.json()


class BaseTextClient(BaseClient):
    """Base class for text clients."""

    def _get_content_from_response(
        self,
        response: requests.Response,
    ) -> io.StringIO:
        return io.StringIO(response.content.decode(response.encoding))



"""Clients module."""

from meteora.clients.aemet import AemetClient
from meteora.clients.agrometeo import AgrometeoClient
from meteora.clients.awel import AWELClient
from meteora.clients.iem import ASOSOneMinIEMClient, METARASOSIEMClient
from meteora.clients.meteocat import MeteocatClient
from meteora.clients.meteoswiss import MeteoSwissClient
from meteora.clients.metoffice import MetOfficeClient
from meteora.clients.netatmo import NetatmoClient
from meteora.clients.noaa import GHCNHourlyClient



"""Tests for Meteora."""

import json
import logging as lg
import os
import tempfile
import unittest
from os import path

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import osmnx as ox
import pandas as pd
import pook
import pytest
import xarray as xr
from pandas.api.types import is_datetime64_any_dtype, is_numeric_dtype

from meteora import qc, settings, utils
from meteora.clients import (
    AemetClient,
    AgrometeoClient,
    ASOSOneMinIEMClient,
    AWELClient,
    GHCNHourlyClient,
    METARASOSIEMClient,
    MeteocatClient,
    MeteoSwissClient,
    MetOfficeClient,
    NetatmoClient,
)
from meteora.mixins import StationsEndpointMixin, VariablesEndpointMixin

tests_dir = "tests"
tests_data_dir = path.join(tests_dir, "data")


def override_settings(module, **kwargs):
    class OverrideSettings:
        def __enter__(self):
            self.old_values = {}
            for key, value in kwargs.items():
                self.old_values[key] = getattr(module, key)
                setattr(module, key, value)

        def __exit__(self, type, value, traceback):
            for key, value in self.old_values.items():
                setattr(module, key, value)

    return OverrideSettings()


class TestUtils(unittest.TestCase):
    def test_geo_utils(self):
        # geo utils
        # dms to dd
        dms_ser = pd.Series(["413120N"])
        dd_ser = utils.dms_to_decimal(dms_ser)
        self.assertTrue(is_numeric_dtype(dd_ser))

    def test_ts_utils(self):
        # time series utils
        # long to wide
        ts_df = pd.read_csv(
            path.join(tests_data_dir, "ts-df.csv"),
            index_col=["station_id", "time"],
            parse_dates=True,
            date_format="%Y-%m-%d %H:%M:%S",
        )
        wide_ts_df = utils.long_to_wide(ts_df)
        # test wide data frame form
        self.assertIsInstance(wide_ts_df.columns, pd.MultiIndex)
        self.assertIsInstance(wide_ts_df.index, pd.DatetimeIndex)
        # with only one variable, we should have only one column level
        wide_ts_df = utils.long_to_wide(ts_df, variables=["temperature"])
        self.assertEqual(len(wide_ts_df.columns.names), 1)
        self.assertIsInstance(wide_ts_df.index, pd.DatetimeIndex)

        # long to cube (xvec)
        stations_gdf = gpd.read_file(
            path.join(tests_data_dir, "stations.gpkg")
        ).set_index(settings.STATIONS_ID_COL)
        with pytest.raises(KeyError):
            # if stations_gdf does not cover all stations in ts_df a KeyError is also
            # raised
            utils.long_to_cube(ts_df, stations_gdf.iloc[:2])
            # attempting to convert from the wide form also raises a KeyError
            utils.wide_to_cube(wide_ts_df, stations_gdf)
        # test proper conversion
        ts_cube = utils.long_to_cube(ts_df, stations_gdf)
        # test an xarray dataset is returned
        self.assertIsInstance(ts_cube, xr.Dataset)
        # test that the time column is in the coordinates
        self.assertIn(ts_df.index.names[1], ts_cube.coords)
        # test that the variable columns are in the data_vars
        self.assertTrue(all([var in ts_cube.data_vars for var in ts_df.columns]))
        # test that it has a dimension with geometry
        self.assertIn("geometry", ts_cube.xvec.geom_coords)
        self.assertIn("geometry", ts_cube.xvec.geom_coords_indexed)
        # test that there is a coordinate with the station ids and that all its values
        # are in the stations_gdf index
        self.assertIn(settings.STATIONS_ID_COL, ts_cube.coords)
        self.assertLessEqual(
            set(ts_cube[settings.STATIONS_ID_COL].values), set(stations_gdf.index)
        )

    def test_meteo_utils(self):
        # meteo utils (heatwave detection)
        ts_df = pd.read_csv(
            path.join(tests_data_dir, "wide-ts-df.csv"),
            index_col="time",
            parse_dates=True,
        )
        # increasing the temperature threshold should result in a less or equal number
        # of heatwave periods
        self.assertLessEqual(
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=27,
                )
            ),
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                )
            ),
        )
        # increasing the duration threshold should result in a less or equal number of
        # heatwave periods
        self.assertLessEqual(
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                    heatwave_n_consecutive_days=3,
                )
            ),
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                    heatwave_n_consecutive_days=2,
                )
            ),
        )
        # using the daily maximum temperature instead of the mean should result in a
        # greater or equal number of heatwave periods
        self.assertGreaterEqual(
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                    heatwave_n_consecutive_days=2,
                    station_agg_func="max",
                )
            ),
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                    heatwave_n_consecutive_days=2,
                    station_agg_func="mean",
                )
            ),
        )
        # using the max instead of the mean to aggregate the stations should result in a
        # greater or equal number of heatwave periods
        self.assertGreaterEqual(
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                    heatwave_n_consecutive_days=2,
                    inter_station_agg_func="max",
                )
            ),
            len(
                utils.get_heatwave_periods(
                    ts_df,
                    heatwave_t_threshold=25,
                    heatwave_n_consecutive_days=2,
                    inter_station_agg_func="mean",
                )
            ),
        )
        # get the time series data for the heatwave periods
        for kwargs in [{}, {"heatwave_t_threshold": 25}]:
            heatwave_ts_df = utils.get_heatwave_ts_df(
                ts_df,
                **kwargs,
            )
            # test that the data frame has a multi-index with the heatwave periods and
            # time as well as the station ids as columns
            self.assertIsInstance(heatwave_ts_df.index, pd.MultiIndex)
            self.assertEqual(
                heatwave_ts_df.index.names,
                [
                    "heatwave",
                    ts_df.index.name,
                ],
            )
        # test that we can also get it from the heatwave periods
        heatwave_periods = utils.get_heatwave_periods(
            ts_df,
        )
        heatwave_ts_df = utils.get_heatwave_ts_df(
            ts_df,
            heatwave_periods=heatwave_periods,
        )
        # test that we have an outermost index with the heatwave periods
        self.assertEqual(
            len(heatwave_ts_df.index.get_level_values("heatwave").unique()),
            len(heatwave_periods),
        )
        # test that an empty time series data frame is returned if no heatwave periods
        # are found
        # # test that a message is logged if no heatwave periods are found
        # with self.assertLogs(settings.LOG_NAME, level=lg.WARNING) as cm:
        #     settings.LOG_CONSOLE = True
        #     self.assertIn("empty", cm.output)
        ts_df = utils.get_heatwave_ts_df(ts_df, heatwave_t_threshold=100)
        self.assertTrue(ts_df.empty)

        # logger
        def test_logging():
            utils.log("test a fake default message")
            utils.log("test a fake debug", level=lg.DEBUG)
            utils.log("test a fake info", level=lg.INFO)
            utils.log("test a fake warning", level=lg.WARNING)
            utils.log("test a fake error", level=lg.ERROR)

        test_logging()
        with override_settings(settings, LOG_CONSOLE=True):
            test_logging()
        with override_settings(settings, LOG_FILE=True):
            test_logging()

        # timestamps
        utils.ts(style="date")
        utils.ts(style="datetime")
        utils.ts(style="time")


def test_region_arg():
    # we will use Agrometeo (since it does not require API keys) to test the region arg
    nominatim_query = "Pully, Switzerland"
    gdf = ox.geocode_to_gdf(nominatim_query)
    with tempfile.TemporaryDirectory() as tmp_dir:
        filepath = path.join(tmp_dir, "foo.gpkg")
        gdf.to_file(filepath)
        for region in [nominatim_query, gdf, filepath]:
            client = AgrometeoClient(region=region)
            stations_gdf = client.stations_gdf
            assert len(stations_gdf) >= 1
    # now test naive geometries without providing CRS, so first ensure that we have them
    # in the same CRS as the client
    gdf = gdf.to_crs(client.CRS)
    for region in [gdf.total_bounds, gdf["geometry"].iloc[0]]:
        client = AgrometeoClient(region=region)
        stations_gdf = client.stations_gdf
        assert len(stations_gdf) >= 1


def test_qc():
    # read a wide ts df
    # ACHTUNG: select only 3 days so that tests run faster (comparison lineplots can be
    # slow)
    ts_df = pd.read_csv(
        path.join(tests_data_dir, "wide-ts-df.csv"), index_col="time", parse_dates=True
    ).iloc[:72]

    # test comparison lineplot
    discard_stations = ts_df.columns[:2]
    # check that there are four lines in the plot (2 mean + 2 CI lines)
    assert len(qc.comparison_lineplot(ts_df, discard_stations).lines) == 4
    # test that if we plot discarded stations individually, we get a line for each
    # discarded station (plus two for the kept ones, i.e., the mean and CI lines)
    assert (
        len(qc.comparison_lineplot(ts_df, discard_stations).lines)
        == len(discard_stations) + 2
    )
    # test that we can plot in a given axis
    fig, ax = plt.subplots()
    assert len(ax.lines) == 0
    qc.comparison_lineplot(ts_df, discard_stations, ax=ax)
    assert len(ax.lines) == 4

    # test mislocated stations
    # generate a random gdf with the same stations as ts_df
    stations_gser = gpd.GeoSeries(
        gpd.points_from_xy(
            np.random.rand(len(ts_df.columns)), np.random.rand(len(ts_df.columns))
        ),
        index=ts_df.columns,
        crs=4326,
    )
    # duplicate some station location
    src_station = ts_df.columns[0]
    dst_station = ts_df.columns[1]
    stations_gser.loc[src_station] = stations_gser.loc[dst_station]
    # test that we get the duplicated stations
    mislocated_stations = qc.get_mislocated_stations(stations_gser)
    for station in [src_station, dst_station]:
        assert station in mislocated_stations

    # test unreliable stations
    unreliable_stations = qc.get_unreliable_stations(ts_df)
    assert len(unreliable_stations) >= 0
    # test threshold (default is 0.2)
    # test that a higher threshold returns at most the same stations
    assert len(qc.get_unreliable_stations(ts_df, unreliable_threshold=0.3)) <= len(
        unreliable_stations
    )
    # test that a lower threshold returns at least the same stations
    assert len(qc.get_unreliable_stations(ts_df, unreliable_threshold=0.1)) >= len(
        unreliable_stations
    )
    # test that we get an empty list if we set the threshold to 1
    assert qc.get_unreliable_stations(ts_df, unreliable_threshold=1 == [])

    # test elevation adjustment
    # generate a random elevation series with stations as index
    station_elevation_ser = pd.Series(
        np.random.rand(len(ts_df.columns)) * 100 + 1, index=ts_df.columns
    )
    # adjust with the default lapse rate (0.0065)
    adj_ts_df = qc.elevation_adjustment(ts_df, station_elevation_ser)
    # test that the adjusted ts_df has the same shape and indexing
    assert adj_ts_df.shape == ts_df.shape
    assert adj_ts_df.index.equals(ts_df.index)
    assert adj_ts_df.columns.equals(ts_df.columns)
    # test that a higher lapse rate increases (strict) the range of the adjusted values
    # technically this may not work with elevations smaller than 1
    high_adj_ts_df = qc.elevation_adjustment(
        ts_df, station_elevation_ser, atmospheric_lapse_rate=0.2
    )
    assert adj_ts_df.min().min() > high_adj_ts_df.min().min()
    assert adj_ts_df.max().max() < high_adj_ts_df.max().max()

    # test outlier detection
    outlier_stations = qc.get_outlier_stations(ts_df)
    assert len(outlier_stations) >= 0
    # test tail range (default high_alpha=0.95, low_alpha=0.01)
    # test that a smaller tail range returns at least the same stations
    assert len(qc.get_outlier_stations(ts_df, low_alpha=0.1, high_alpha=0.9)) >= len(
        outlier_stations
    )
    # test that a bigger tail range returns at most the same stations
    assert len(qc.get_outlier_stations(ts_df, low_alpha=0.01, high_alpha=0.99)) <= len(
        outlier_stations
    )
    # test station outlier threshold (default 0.2)
    # test that a higher outlier threshold returns at most the same stations
    station_outlier_threshold = 0.3
    assert len(
        qc.get_outlier_stations(
            ts_df, station_outlier_threshold=station_outlier_threshold
        )
    ) <= len(outlier_stations)
    # test that a lower outlier threshold returns at least the same stations
    assert len(
        qc.get_outlier_stations(
            ts_df, station_outlier_threshold=station_outlier_threshold
        )
    ) >= len(outlier_stations)

    # test indoor station detection
    indoor_stations = qc.get_indoor_stations(ts_df)
    assert len(indoor_stations) >= 0
    # test correlation threshold (default 0.9)
    # test that a higher threshold returns at least the same stations
    assert len(
        qc.get_indoor_stations(ts_df, station_indoor_corr_threshold=0.95)
    ) >= len(indoor_stations)
    # test that a lower threshold returns at most the same stations
    assert len(
        qc.get_indoor_stations(ts_df, station_indoor_corr_threshold=0.85)
    ) <= len(indoor_stations)


class BaseClientTest:
    client_cls = None
    region = None
    variables = ["temperature", "pressure"]
    variable_codes = None
    ts_df_args = None
    ts_df_kwargs = None

    def setUp(self):
        self.client = self.client_cls(region=self.region)

    def test_attributes(self):
        for attr in ["X_COL", "Y_COL", "CRS"]:
            self.assertTrue(hasattr(self.client, attr))
            self.assertIsNotNone(getattr(self.client, attr))

    def test_stations(self):
        if isinstance(self.client, StationsEndpointMixin):
            stations_gdf = self.client.stations_gdf
            assert len(stations_gdf) >= 1
            self.assertEqual(stations_gdf.index.name, settings.STATIONS_ID_COL)

    def test_variables(self):
        if isinstance(self.client, VariablesEndpointMixin):
            variables_df = self.client.variables_df
            assert len(variables_df) >= 1

    def test_time_series(self):
        if self.ts_df_args is None:
            ts_df_args = []
        else:
            ts_df_args = self.ts_df_args.copy()
        if self.ts_df_kwargs is None:
            ts_df_kwargs = {}
        else:
            ts_df_kwargs = self.ts_df_kwargs.copy()
        for variables in [self.variables, self.variable_codes]:
            ts_df = self.client.get_ts_df(self.variables, *ts_df_args, **ts_df_kwargs)
            # test data frame shape
            assert len(ts_df.columns) == len(self.variables)
            # TODO: use "station" as `level` arg?
            # ACHTUNG: using the <= because in many cases some stations are listed in
            # the stations endpoint but do not return data (maybe inactive?)
            assert len(ts_df.index.get_level_values(0).unique()) <= len(
                self.client.stations_gdf
            )
            # TODO: use "time" as `level` arg?
            assert is_datetime64_any_dtype(ts_df.index.get_level_values(1))
            # test that index is sorted (note that we need to test it as a multi-index
            # because otherwise the time index alone is not unique in long data frames
            assert ts_df.index.is_monotonic_increasing
            # test index labels
            assert ts_df.index.names == [settings.STATIONS_ID_COL, settings.TIME_COL]


class APIKeyClientTest(BaseClientTest):
    stations_response_file = None

    def setUp(self):
        self.client = self.client_cls(self.region, self.api_key)

    def test_attributes(self):
        super().test_attributes()
        self.assertTrue(hasattr(self.client, "_api_key"))
        self.assertIsNotNone(self.client._api_key)


class APIKeyHeaderClientTest(APIKeyClientTest):
    def test_attributes(self):
        super().test_attributes()
        self.assertTrue("X-API-KEY" in self.client.request_headers)
        self.assertIsNotNone(self.client.request_headers["X-API-KEY"])


class APIKeyParamClientTest(APIKeyClientTest):
    def test_attributes(self):
        super().test_attributes()
        self.assertTrue(hasattr(self.client, "_api_key_param_name"))
        api_key_param_name = self.client._api_key_param_name
        self.assertTrue(api_key_param_name in self.client.request_params)
        self.assertIsNotNone(self.client.request_params[api_key_param_name])


class OAuth2ClientTest(BaseClientTest):
    def setUp(self):
        self.client = self.client_cls(
            self.region, self.client_id, self.client_secret, token=self.token
        )


class AemetClientTest(APIKeyParamClientTest, unittest.TestCase):
    client_cls = AemetClient
    region = "Catalunya"
    api_key = os.getenv("AEMET_API_KEY", "")
    # ACHTUNG: the test data that we have for AEMET does NOT include the "pressure"
    # variable even though this is a listed variable in the AEMET API, which probably
    # means that the test data stations do NOT have that variable - TODO: raise an
    # informative error if the variable is not available for the specified stations
    variables = ["temperature", "precipitation"]
    variable_codes = ["ta", "prec"]

    @pook.on
    def test_all(self):
        # test stations, variables and time series in the same method because we need
        # to mock the same requests
        with open(path.join(tests_data_dir, "aemet-stations.json")) as src:
            response_dict = json.load(src)
            pook.get(
                f"{AemetClient._stations_endpoint}?api_key={self.api_key}",
                response_json=response_dict,
                persist=True,
            )
        with open(path.join(tests_data_dir, "aemet-stations-datos.json")) as src:
            pook.get(response_dict["datos"], response_json=json.load(src), persist=True)
        super().test_stations()

        with open(path.join(tests_data_dir, "aemet-var-ts.json")) as src:
            response_dict = json.load(src)
            pook.get(
                f"{AemetClient._variables_endpoint}?api_key={self.api_key}",
                response_json=response_dict,
                persist=True,
            )
        with open(path.join(tests_data_dir, "aemet-var-ts-metadatos.json")) as src:
            pook.get(
                response_dict["metadatos"],
                response_json=json.load(src),
            )

        with open(path.join(tests_data_dir, "aemet-var-ts-datos.json")) as src:
            pook.get(
                response_dict["datos"],
                response_json=json.load(src),
            )

        # test stations
        super().test_stations()

        # test variables
        # variables_df = self.client.variables_df
        assert len(self.client.variables_df) >= 1

        # test time series
        self.client.get_ts_df(self.variables)
        # ACHTUNG: for some reason (internal to Aemet's API), we get more stations
        # from the stations endpoint than from the time series endpoint, so the
        # assertions of the `test_time_series` method would fail
        # super().test_time_series()

    # the tests are already done in `test_all` so we override it here with empty methods
    def test_stations(self):
        pass

    def test_variables(self):
        pass

    def test_time_series(self):
        pass


class AgrometeoClientTest(BaseClientTest, unittest.TestCase):
    client_cls = AgrometeoClient
    region = "Pully, Switzerland"
    variable_codes = [1, 18]
    start_date = "2022-03-22"
    end_date = "2022-03-23"
    ts_df_args = [start_date, end_date]


class AWELClientTest(BaseClientTest, unittest.TestCase):
    client_cls = AWELClient
    region = "Zürich, Switzerland"
    variables = ["temperature", "relative_humidity"]
    variable_codes = ["temperature", "humidity"]
    start_date = "2022-03-22"
    end_date = "2022-03-23"
    ts_df_args = [start_date, end_date]


class IEMBaseClientTest(BaseClientTest):
    region = "Vermont"
    start_date = "2022-03-22"
    end_date = "2022-03-23"
    ts_df_args = [start_date, end_date]


class ASOSOneMinIEMClientTest(IEMBaseClientTest, unittest.TestCase):
    client_cls = ASOSOneMinIEMClient
    variable_codes = ["tmpf", "pres1"]


class METARASOSSIEMClientTest(IEMBaseClientTest, unittest.TestCase):
    client_cls = METARASOSIEMClient
    variable_codes = ["tmpf", "mslp"]


class MeteocatClientTest(APIKeyHeaderClientTest, unittest.TestCase):
    client_cls = MeteocatClient
    region = "Conca de Barberà"
    api_key = os.environ["METEOCAT_API_KEY"]
    variable_codes = [32, 34]
    start_date = "2022-03-22"
    end_date = "2022-03-23"
    ts_df_args = [start_date, end_date]


class MeteoSwissClientTest(BaseClientTest, unittest.TestCase):
    client_cls = MeteoSwissClient
    region = "Pully, Switzerland"
    variable_codes = ["tre200s0", "prestas0"]
    start_date = "2022-03-22"
    end_date = "2022-03-23"
    ts_df_args = [start_date, end_date]


class MetOfficeClientTest(APIKeyParamClientTest, unittest.TestCase):
    client_cls = MetOfficeClient
    region = "Edinburgh"
    api_key = os.environ["METOFFICE_API_KEY"]
    variable_codes = ["T", "P"]


class NetatmoClientTest(OAuth2ClientTest, unittest.TestCase):
    client_cls = NetatmoClient
    region = "Passanant i Belltall"
    client_id = os.environ["NETATMO_CLIENT_ID"]
    client_secret = os.environ["NETATMO_CLIENT_SECRET"]
    token = {"access_token": os.environ["NETATMO_ACCESS_TOKEN"]}
    # token = None
    variables = ["temperature", "relative_humidity"]
    variable_codes = ["temperature", "humidity"]
    start_date = "2024-12-22"
    end_date = "2024-12-23"
    ts_df_args = [start_date, end_date]

    # def setUp(self):
    #     with requests_mock.Mocker() as m:
    #         m.post(
    #             netatmo.OAUTH2_TOKEN_ENDPOINT,
    #             json={"token_type": "bearer", "access_token": "abcd"},
    #         )
    #         super().setUp()

    @pook.on
    def test_stations(self):
        with open(path.join(tests_data_dir, "netatmo-stations.json")) as src:
            pook.get(
                "https://api.netatmo.com/api/getpublicdata?lon_sw=1.1635994&"
                "lat_sw=41.48811&lon_ne=1.2635994000000002&lat_ne=41.58811",
                response_json=json.load(src),
            )
        super().test_stations()

    @pook.on
    def test_time_series(self):
        with open(path.join(tests_data_dir, "netatmo-time-series.json")) as src:
            pook.get(
                "https://api.netatmo.com/api/getmeasure?type=temperature%2Chumidity"
                "&scale=30min&limit=1024&optimize=True&real_time=False&"
                "device_id=70%3Aee%3A50%3A74%3A2a%3Aba&"
                "module_id=02%3A00%3A00%3A73%3Ae0%3A7e&"
                "date_begin=1734825600.0&date_end=1734912000.0",
                response_json=json.load(src),
            )
        super().test_time_series()


class GHCNHourlyClientTest(BaseClientTest, unittest.TestCase):
    client_cls = GHCNHourlyClient
    region = "Pully, Switzerland"
    variable_codes = ["temperature", "relative_humidity"]
    start_date = "2022-03-22"
    end_date = "2022-03-23"
    ts_df_args = [start_date, end_date]



"""Settings."""

import logging as lg

import requests_cache

# core
STATIONS_ID_COL = "station_id"
TIME_COL = "time"
SJOIN_KWARGS = {"how": "inner", "predicate": "intersects"}

## netatmo
NETATMO_ON_GET_ERROR = "log"  # or "raise"

# qc
ATMOSPHERIC_LAPSE_RATE = 0.0065
OUTLIER_LOW_ALPHA = 0.01
OUTLIER_HIGH_ALPHA = 0.95
STATION_OUTLIER_THRESHOLD = 0.2
STATION_INDOOR_CORR_THRESHOLD = 0.9
UNRELIABLE_THRESHOLD = 0.2

# utils
## meteo
HEATWAVE_T_THRESHOLD = 25
HEATWAVE_N_CONSECUTIVE_DAYS = 3
HEATWAVE_STATION_AGG_FUNC = "mean"
HEATWAVE_INTER_STATION_AGG_FUNC = "mean"

REQUEST_KWARGS = {}
# PAUSE = 1
ERROR_PAUSE = 60
# TIMEOUT = 180
## cache
USE_CACHE = True
CACHE_NAME = "meteora-cache"
CACHE_BACKEND = "sqlite"
CACHE_EXPIRE = requests_cache.NEVER_EXPIRE

## logging
LOG_CONSOLE = False
LOG_FILE = False
LOG_FILENAME = "meteora"
LOG_LEVEL = lg.INFO
LOG_NAME = "meteora"
LOGS_FOLDER = "./logs"



"""Authentication mixins."""

from abc import ABC, abstractmethod

from meteora.utils import abstract_attribute


class APIKeyMixin(ABC):
    """API key mixin."""

    @abstract_attribute
    def _api_key(self) -> str:
        # https://stackoverflow.com/questions/23831510/abstract-attribute-not-property
        pass


class APIKeyHeaderMixin(APIKeyMixin):
    """API key as request header mixin."""

    @property
    def request_headers(self) -> dict:
        """Request headers."""
        try:
            return self._request_headers
        except AttributeError:
            self._request_headers = {"X-API-KEY": self._api_key}
            return self._request_headers


class APIKeyParamMixin(APIKeyMixin):
    """API key as request parameter mixin."""

    @property
    @abstractmethod
    def _api_key_param_name(self) -> str:
        pass

    @property
    def request_params(self) -> dict:
        """Request parameters."""
        try:
            return self._request_params
        except AttributeError:
            self._request_params = {self._api_key_param_name: self._api_key}
            return self._request_params



"""Variables mixins."""

import pandas as pd

from meteora.utils import VariablesType, abstract_attribute

# https://public.wmo.int/en/programmes/global-climate-observing-system/essential-climate-variables
ECVS = [
    "precipitation",
    "pressure",
    "surface_radiation_longwave",
    "surface_radiation_shortwave",
    "surface_wind_speed",
    "surface_wind_direction",
    "temperature",
    "water_vapour",
]


class VariablesMixin:
    """Variables Mixin."""

    @abstract_attribute
    def _variables_id_col(self) -> str:
        pass

    @abstract_attribute
    def _ecv_dict(self) -> dict:
        pass

    def _process_variable_arg(self, variable: str | int) -> str | int:
        # process the variable arg
        # variable is a string that can be either:
        # a) a variable code according to the provider's nomenclature
        # b) an essential climate variable (ECV) following the Meteora nomenclature
        if isinstance(variable, int) or variable.isdigit():
            # case a: if variable is an integer, assert that it is a valid variable code
            variable_id = int(variable)
            if variable_id not in self.variables_df[self._variables_id_col].values:
                raise ValueError(f"variable {variable} is not a valid variable id")
        elif variable in self.variables_df[self._variables_id_col].values:
            # still case a: if variable is a variable code, but it is a string - then,
            # just return it as it is
            return variable
        else:
            # case b: if variable is an ECV, it will be a key in the ECV_DICT so
            # the provider's variable code can be retrieved directly, otherwise we
            # assume that variable is a variable name (provider's nomenclature).
            variable_id = self._ecv_dict.get(variable, variable)
            # variable_code = self.variables_df.loc[
            #     self.variables_df[self._variables_label_col] == variable_name,
            #     self._variables_code_col,
            # ].item()

        return variable_id

    def _get_variable_id_ser(self, variables: VariablesType) -> pd.Series:
        """Given the `variables` argument, return a list of variable codes."""
        if not pd.api.types.is_list_like(variables):
            variables = [variables]
        # ensure variable codes have the same dtype as in the variables data frame
        return pd.Series(
            [self._process_variable_arg(variable) for variable in variables],
            index=variables,
            dtype=self.variables_df[self._variables_id_col].dtype,
        )


class VariablesHardcodedMixin(VariablesMixin):
    """Hardcoded variables mixin."""

    @abstract_attribute
    def _variables_dict(self) -> dict:
        pass

    @abstract_attribute
    def _variables_label_col(self) -> str:
        pass

    @property
    def variables_df(self) -> pd.DataFrame:
        """Variables dataframe."""
        try:
            return self._variables_df
        except AttributeError:
            variables_df = pd.DataFrame(
                self._variables_dict.items(),
                columns=[self._variables_id_col, self._variables_label_col],
            )
            self._variables_df = variables_df
            return self._variables_df


class VariablesEndpointMixin(VariablesMixin):
    """Variables endpoint mixin."""

    @abstract_attribute
    def _variables_endpoint(self) -> str:
        pass

    @property
    def variables_df(self) -> pd.DataFrame:
        """Variables dataframe."""
        try:
            return self._variables_df
        except AttributeError:
            response_content = self._get_content_from_url(self._variables_endpoint)
            self._variables_df = self._variables_df_from_content(response_content)
            return self._variables_df
